######################################################################
########### THIS FILE IS GENERATED, DO NOT EDIT MANUALLY!!! ##########
########### THIS FILE IS GENERATED, DO NOT EDIT MANUALLY!!! ##########
########### THIS FILE IS GENERATED, DO NOT EDIT MANUALLY!!! ##########
########### THIS FILE IS GENERATED, DO NOT EDIT MANUALLY!!! ##########
########### THIS FILE IS GENERATED, DO NOT EDIT MANUALLY!!! ##########
######################################################################


from wunderkafka import librdkafka


if librdkafka.__version__ >= (2, 0, 0):
    from wunderkafka.config.generated.v2_0_0.fields import *  # type: ignore[assignment]
if librdkafka.__version__ >= (2, 1, 0):
    from wunderkafka.config.generated.v2_1_0.fields import *  # type: ignore[assignment]
if librdkafka.__version__ >= (2, 2, 0):
    from wunderkafka.config.generated.v2_2_0.fields import *  # type: ignore[assignment]
if librdkafka.__version__ >= (2, 3, 0):
    from wunderkafka.config.generated.v2_3_0.fields import *  # type: ignore[assignment]
if librdkafka.__version__ >= (2, 4, 0):
    from wunderkafka.config.generated.v2_4_0.fields import *  # type: ignore[assignment]
if librdkafka.__version__ >= (2, 5, 0):
    from wunderkafka.config.generated.v2_5_0.fields import *  # type: ignore[assignment]
if librdkafka.__version__ >= (2, 6, 0):
    from wunderkafka.config.generated.v2_6_0.fields import *  # type: ignore[assignment]
if librdkafka.__version__ >= (2, 7, 0):
    from wunderkafka.config.generated.v2_7_0.fields import *  # type: ignore[assignment]
if librdkafka.__version__ >= (2, 8, 0):
    from wunderkafka.config.generated.v2_8_0.fields import *  # type: ignore[assignment]
if librdkafka.__version__ >= (2, 9, 0):
    from wunderkafka.config.generated.v2_9_0.fields import *  # type: ignore[assignment]
if librdkafka.__version__ >= (2, 10, 0):
    from wunderkafka.config.generated.v2_10_0.fields import *  # type: ignore[assignment]
if librdkafka.__version__ >= (2, 11, 0):
    from wunderkafka.config.generated.v2_11_0.fields import *  # type: ignore[assignment]
if librdkafka.__version__ >= (2, 12, 0):
    from wunderkafka.config.generated.v2_12_0.fields import *  # type: ignore[assignment]



import os
from typing import Any
from unittest.mock import patch

from aws_msk_iam_sasl_signer import MSKAuthTokenProvider

from wunderkafka.time import ts2dt
from wunderkafka.logger import logger

Token = str
ExpiryMs = int


class OAuthCallback:
    def __init__(self, session_name: str):
        self.session_name = session_name

    def __call__(self, _: Any) -> tuple[Token, ExpiryMs]:  # noqa: ANN401
        with patch.dict(os.environ, {"AWS_ROLE_SESSION_NAME": self.session_name}):
            return oauth_callback(None)


# as the librdkafka on `sasl.oauthbearer.config` doc says,
#   "The format is implementation-dependent and must be parsed accordingly"
# so we can't annotate `_` with something meaningful.
# we just don't use it for now
def oauth_callback(_: Any) -> tuple[Token, ExpiryMs]:  # noqa: ANN401
    region = os.getenv("AWS_REGION") or os.getenv("AWS_DEFAULT_REGION")
    if not region:
        msg = "Region not set - export AWS_REGION or AWS_DEFAULT_REGION before running."
        raise RuntimeError(msg)
    auth_token, expiry_ms = MSKAuthTokenProvider.generate_auth_token(region)
    # Note that this library expects oauth_cb to return expiry time in seconds since epoch,
    # while the token generator returns expiry in ms
    logger.info(f"Generated auth token for region {region} with expiry {ts2dt(expiry_ms)}")
    return auth_token, expiry_ms / 1000



from typing import Any, Union, Optional

from wunderkafka.types import DeliveryCallback
from wunderkafka.message import MessageProtocol as Message
from wunderkafka.tests.message import Message as MessageStub
from wunderkafka.producers.bytes import BytesProducer
from wunderkafka.callbacks.producer import error_callback


class TestProducer(BytesProducer):
    def __init__(self) -> None:
        self.sent: list[Message] = []

    def send_message(
        self,
        topic: str,
        value: Optional[Union[str, bytes]] = None,
        key: Optional[Union[str, bytes]] = None,
        partition: Optional[int] = None,
        on_delivery: Optional[DeliveryCallback] = error_callback,
        *args: Any,
        blocking: bool = False,
        **kwargs: Any,
    ) -> None:
        message = MessageStub(topic=topic, value=value, key=key)
        self.sent.append(message)



from typing import Any, Optional
from dataclasses import asdict, is_dataclass

from pydantic import BaseModel

from wunderkafka.compat import ParamSpec
from wunderkafka.serdes.abc import AbstractSerializer, AbstractDescriptionStore
from wunderkafka.serdes.avro import FastAvroSerializer

P = ParamSpec("P")


class AvroModelSerializer(AbstractSerializer):
    def __init__(self, store: Optional[AbstractDescriptionStore] = None) -> None:
        self._serializer = FastAvroSerializer()
        self.store = store

    def serialize(
        self,
        schema: str,
        payload: Any,
        header: Optional[bytes] = None,
        *args: P.args,  # type: ignore[valid-type]
        **kwargs: P.kwargs,  # type: ignore[valid-type]
    ) -> bytes:
        if isinstance(payload, BaseModel):
            dct = payload.model_dump()
        else:
            dct = asdict(payload) if is_dataclass(payload) else dict(payload)  # type: ignore[arg-type]
        return self._serializer.serialize(schema, dct, header)



import struct

from wunderkafka.errors import DeserializerException
from wunderkafka.serdes.abc import AbstractProtocolHandler
from wunderkafka.structures import SRMeta, ParsedHeader
from wunderkafka.serdes.vendors import Actions
from wunderkafka.serdes.protocols import MIN_HEADER_SIZE, get_protocol

# TODO (tribunsky.kir): Get rid of AbstractProtocolHandler?
#                       https://github.com/


class ConfluentClouderaHeadersHandler(AbstractProtocolHandler):
    def parse(self, blob: bytes) -> ParsedHeader:
        if len(blob) < MIN_HEADER_SIZE:
            msg = f"Message is too small to decode: ({blob!r})"
            raise DeserializerException(msg)

        # 1st byte is magic byte.
        [protocol_id] = struct.unpack(">b", blob[0:1])

        protocol = get_protocol(protocol_id, Actions.deserialize)

        # already read 1 byte from header as protocol id
        meta = struct.unpack(protocol.mask.unpack, blob[1 : 1 + protocol.header_size])

        if protocol_id == 1:
            schema_id = None
            schema_meta_id, schema_version = meta
        else:
            [schema_id] = meta
            schema_meta_id = None
            schema_version = None

        return ParsedHeader(
            protocol_id=protocol_id,
            meta_id=schema_meta_id,
            schema_id=schema_id,
            schema_version=schema_version,
            size=protocol.header_size + 1,
        )

    def pack(self, protocol_id: int, meta: SRMeta) -> bytes:
        protocol = get_protocol(protocol_id, Actions.serialize)

        if protocol_id == 1:
            if meta.meta_id is None:
                err_msg = f"No meta id for protocol_id {protocol_id}. Please, check response from Schema Registry."
                raise ValueError(err_msg)
            return struct.pack(protocol.mask.pack, protocol_id, meta.meta_id, meta.schema_version)
        return struct.pack(protocol.mask.pack, protocol_id, meta.schema_id)



import io
from json import loads
from typing import Any, Optional

from fastavro import parse_schema, schemaless_writer

from wunderkafka.serdes.abc import AbstractSerializer, AbstractDescriptionStore
from wunderkafka.serdes.avro.types import FastAvroParsedSchema


class FastAvroSerializer(AbstractSerializer):
    def __init__(self, store: Optional[AbstractDescriptionStore] = None) -> None:
        self._cache: dict[str, FastAvroParsedSchema] = {}
        self.store = store

    def serialize(
        self,
        schema: str,
        payload: Any,
        header: Optional[bytes] = None,
        topic: Optional[str] = None,
        *,
        is_key: Optional[bool] = None,
    ) -> bytes:
        if schema not in self._cache:
            self._cache[schema] = parse_schema(loads(schema))
        writer_schema = self._cache[schema]
        with io.BytesIO() as buffer:
            if header is not None:
                buffer.write(header)
            schemaless_writer(buffer, writer_schema, payload)
            return buffer.getvalue()



######################################################################
########### THIS FILE IS GENERATED, DO NOT EDIT MANUALLY!!! ##########
########### THIS FILE IS GENERATED, DO NOT EDIT MANUALLY!!! ##########
########### THIS FILE IS GENERATED, DO NOT EDIT MANUALLY!!! ##########
########### THIS FILE IS GENERATED, DO NOT EDIT MANUALLY!!! ##########
########### THIS FILE IS GENERATED, DO NOT EDIT MANUALLY!!! ##########
######################################################################

# Why so: not all configuration parameters of librdkafka may be easily replaced from '_' to '.',
#   therefore, we can't convert on-the-fly from  `ssl_ca` without errors
#   and we don't want to have a nice whitelist, which is arguable
COMMON_FIELDS = (
    "allow.auto.create.topics",
    "api.version.fallback.ms",
    "api.version.request",
    "api.version.request.timeout.ms",
    "background_event_cb",
    "bootstrap.servers",
    "broker.address.family",
    "broker.address.ttl",
    "broker.version.fallback",
    "builtin.features",
    "client.dns.lookup",
    "client.id",
    "client.rack",
    "closesocket_cb",
    "connect_cb",
    "connections.max.idle.ms",
    "debug",
    "default_topic_conf",
    "enable.metrics.push",
    "enable.random.seed",
    "enable.sasl.oauthbearer.unsecure.jwt",
    "enable.ssl.certificate.verification",
    "enabled_events",
    "error_cb",
    "interceptors",
    "internal.termination.signal",
    "log.connection.close",
    "log.queue",
    "log.thread.name",
    "log_level",
    "logger",
    "max.in.flight",
    "max.in.flight.requests.per.connection",
    "message.copy.max.bytes",
    "message.max.bytes",
    "metadata.broker.list",
    "metadata.max.age.ms",
    "oauth_cb",
    "oauthbearer_token_refresh_cb",
    "opaque",
    "open_cb",
    "plugin.library.paths",
    "receive.message.max.bytes",
    "reconnect.backoff.max.ms",
    "reconnect.backoff.ms",
    "resolve_cb",
    "retry.backoff.max.ms",
    "retry.backoff.ms",
    "sasl.kerberos.keytab",
    "sasl.kerberos.kinit.cmd",
    "sasl.kerberos.min.time.before.relogin",
    "sasl.kerberos.principal",
    "sasl.kerberos.service.name",
    "sasl.mechanism",
    "sasl.mechanisms",
    "sasl.oauthbearer.client.id",
    "sasl.oauthbearer.client.secret",
    "sasl.oauthbearer.config",
    "sasl.oauthbearer.extensions",
    "sasl.oauthbearer.method",
    "sasl.oauthbearer.scope",
    "sasl.oauthbearer.token.endpoint.url",
    "sasl.password",
    "sasl.username",
    "security.protocol",
    "socket.connection.setup.timeout.ms",
    "socket.keepalive.enable",
    "socket.max.fails",
    "socket.nagle.disable",
    "socket.receive.buffer.bytes",
    "socket.send.buffer.bytes",
    "socket.timeout.ms",
    "socket_cb",
    "ssl.ca.certificate.stores",
    "ssl.ca.location",
    "ssl.ca.pem",
    "ssl.certificate.location",
    "ssl.certificate.pem",
    "ssl.certificate.verify_cb",
    "ssl.cipher.suites",
    "ssl.crl.location",
    "ssl.curves.list",
    "ssl.endpoint.identification.algorithm",
    "ssl.engine.id",
    "ssl.key.location",
    "ssl.key.password",
    "ssl.key.pem",
    "ssl.keystore.location",
    "ssl.keystore.password",
    "ssl.providers",
    "ssl.sigalgs.list",
    "ssl_ca",
    "ssl_certificate",
    "ssl_engine_callback_data",
    "ssl_key",
    "statistics.interval.ms",
    "stats_cb",
    "throttle_cb",
    "topic.blacklist",
    "topic.metadata.propagation.max.ms",
    "topic.metadata.refresh.fast.interval.ms",
    "topic.metadata.refresh.interval.ms",
    "topic.metadata.refresh.sparse",
)
CONSUMER_FIELDS = (
    "auto.commit.interval.ms",
    "auto.offset.reset",
    "check.crcs",
    "consume.callback.max.messages",
    "consume_cb",
    "coordinator.query.interval.ms",
    "enable.auto.commit",
    "enable.auto.offset.store",
    "enable.partition.eof",
    "fetch.error.backoff.ms",
    "fetch.max.bytes",
    "fetch.message.max.bytes",
    "fetch.min.bytes",
    "fetch.queue.backoff.ms",
    "fetch.wait.max.ms",
    "group.id",
    "group.instance.id",
    "group.protocol",
    "group.protocol.type",
    "group.remote.assignor",
    "heartbeat.interval.ms",
    "isolation.level",
    "max.partition.fetch.bytes",
    "max.poll.interval.ms",
    "offset_commit_cb",
    "partition.assignment.strategy",
    "queued.max.messages.kbytes",
    "queued.min.messages",
    "rebalance_cb",
    "session.timeout.ms",
)
PRODUCER_FIELDS = (
    "acks",
    "batch.num.messages",
    "batch.size",
    "compression.codec",
    "compression.level",
    "compression.type",
    "delivery.report.only.error",
    "delivery.timeout.ms",
    "dr_cb",
    "dr_msg_cb",
    "enable.gapless.guarantee",
    "enable.idempotence",
    "linger.ms",
    "message.send.max.retries",
    "message.timeout.ms",
    "msg_order_cmp",
    "partitioner",
    "partitioner_cb",
    "queue.buffering.backpressure.threshold",
    "queue.buffering.max.kbytes",
    "queue.buffering.max.messages",
    "queue.buffering.max.ms",
    "queuing.strategy",
    "request.required.acks",
    "request.timeout.ms",
    "retries",
    "sticky.partitioning.linger.ms",
    "transaction.timeout.ms",
    "transactional.id",
)



from dataclasses import dataclass


class Mask:
    def __init__(self, value: str):
        self._value = value
        self.__unpack = f">{self._value}"
        self.__pack = f">b{self._value}"

    @property
    def unpack(self) -> str:
        return self.__unpack

    @property
    def pack(self) -> str:
        return self.__pack


@dataclass(frozen=True)
class Protocol:
    header_size: int
    mask: Mask



from typing import Any, Optional

from confluent_kafka.serialization import StringSerializer as _StringSerializer

from wunderkafka.serdes.abc import AbstractSerializer, AbstractDescriptionStore
from wunderkafka.serdes.store import StringRepo


class StringSerializer(AbstractSerializer):
    def __init__(self, store: Optional[AbstractDescriptionStore] = None) -> None:
        self.store = store or StringRepo()
        self._serializer = _StringSerializer()

    def serialize(
        self,
        schema: str,
        payload: Any,
        header: Optional[bytes] = None,
        topic: Optional[str] = None,
        *,
        is_key: Optional[bool] = None,
    ) -> bytes:
        return self._serializer(payload, None)



from __future__ import annotations

from typing import Any, Optional

from wunderkafka.serdes.abc import AbstractSerializer
from wunderkafka.serdes.store import StringRepo


class SchemaLessJSONModelSerializer(AbstractSerializer):
    def __init__(self, **pydantic_kwargs: Any) -> None:
        self.store = StringRepo()
        self._serialization_kwargs = pydantic_kwargs

    def serialize(
        self,
        schema: str,
        obj: Any,
        header: Optional[bytes] = None,
        topic: Optional[str] = None,
        *,
        is_key: Optional[bool] = None,
    ) -> bytes:
        return obj.model_dump_json(**self._serialization_kwargs).encode()



from wunderkafka.schema_registry.cache import SimpleCache
from wunderkafka.schema_registry.transport import KerberizableHTTPClient
from wunderkafka.schema_registry.clients.cloudera import ClouderaSRClient
from wunderkafka.schema_registry.clients.confluent import ConfluentSRClient

__all__ = [
    "SimpleCache",
    "KerberizableHTTPClient",
    "ClouderaSRClient",
    "ConfluentSRClient",
]



from typing import Final

from wunderkafka.serdes.vendors import get_protocol, get_protocols

MIN_HEADER_SIZE: Final[int] = min([pr.header_size for pr in get_protocols()]) + 1


__all__ = [
    "MIN_HEADER_SIZE",
    "get_protocol",
]



"""This module contains some predefined callbacks to interact with librdkafka."""

from __future__ import annotations

from confluent_kafka import TopicPartition

from wunderkafka.logger import logger
from wunderkafka.structures import Timestamp
from wunderkafka.consumers.abc import AbstractConsumer


def reset_partitions(consumer: AbstractConsumer, partitions: list[TopicPartition]) -> None:
    """
    Set a specific offset for assignment after subscription.

    Depending on the type of subscription, will set offset or timestamp.

    :param consumer:            Consumer, which is subscribed to topics.
    :param partitions:          List of TopicPartitions, which is returned from the underlying library.
    """
    new_offsets = consumer.subscription_offsets
    if new_offsets is None:
        logger.warning(
            f"{consumer}: re-assigned (using auto.offset.reset={consumer.config.auto_offset_reset})",
        )
        return
    by_offset = []
    by_ts = []
    for partition in partitions:
        new_offset = new_offsets[partition.topic]
        if new_offset is None:
            by_offset.append(partition)
        else:
            partition.offset = new_offset.value
            if isinstance(new_offset, Timestamp):
                logger.info(f"Setting {new_offset}...")
                by_ts.append(partition)
            else:
                by_offset.append(partition)
    if by_ts:
        by_ts = consumer.offsets_for_times(by_ts)
    new_partitions = by_ts + by_offset
    consumer.assign(new_partitions)
    logger.info(f"{consumer} assigned to {new_partitions}")
    consumer.subscription_offsets = None



import time
import subprocess

from confluent_kafka import KafkaError

from wunderkafka.logger import logger
from wunderkafka.config.rdkafka import RDKafkaConfig
from wunderkafka.config.generated import enums

REQUIRES_KERBEROS = frozenset([enums.SecurityProtocol.sasl_ssl, enums.SecurityProtocol.sasl_plaintext])


def exclude_gssapi(builtin_features: str) -> str:
    features = [feature.strip() for feature in builtin_features.split(",") if feature.strip() != "sasl_gssapi"]
    return ", ".join(features)


def config_requires_kerberos(config: RDKafkaConfig) -> bool:
    if config.sasl_mechanism.casefold() != "GSSAPI".casefold():
        return False
    if config.sasl_kerberos_keytab is None:
        return False
    return config.security_protocol in {enums.SecurityProtocol.sasl_ssl, enums.SecurityProtocol.sasl_plaintext}


def challenge_krb_arg(exc: KafkaError, config: RDKafkaConfig) -> RDKafkaConfig:
    """
    Check if we can just skip kerberos configuration which comes to RDKafkaConfig from documentation default.

    Currently, I didn't find anything in confluent-kafka to query `builtin.features` from python,
    so we are just checking error while instantiating original consumer/producer
    and override corresponding config values.
    """
    logger.warning("Error while instantiating consumer/producer. Checking builtin.features...")
    # https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#kafkaexception
    error = exc.args[0]
    if error.code() != KafkaError._INVALID_ARG:
        raise
    if "sasl_gssapi" not in error.str():
        raise
    if config_requires_kerberos(config):
        raise
    else:
        logger.warning(
            " ".join(
                [
                    "Looks like that current client configuration doesn't require kerberos.",
                    "As it is not supported by currently installed build, skipping this option.",
                ]
            )
        )
        old = config.builtin_features
        new = exclude_gssapi(config.builtin_features)
        logger.warning(f"Changing builtin.features: {old} -> {new}")
        config.builtin_features = new
        return config


def init_kerberos(kinit_cmd: str, timeout: int = 60) -> None:
    t0 = time.perf_counter()
    refresh_cmd = kinit_cmd.split()
    try:
        subprocess.run(refresh_cmd, timeout=timeout, stdout=subprocess.PIPE, check=True)
    # Will retry shortly
    except (subprocess.CalledProcessError, subprocess.TimeoutExpired) as exc:
        logger.error(exc.output)
        logger.error(exc.stdout)
        logger.error(exc.stderr)
        logger.error(f"Command: {refresh_cmd} exit error: {str(exc)}")
        logger.warning("Krb not refreshed!")
    else:
        duration = int(1000 * (time.perf_counter() - t0))
        logger.info(f"Refreshed! ({duration} ms)")



import json
from typing import Any, Optional

from wunderkafka.serdes.abc import AbstractSerializer
from wunderkafka.serdes.store import StringRepo


class SchemaLessJSONSerializer(AbstractSerializer):
    def __init__(self) -> None:
        self.store = StringRepo()

    def serialize(
        self,
        schema: str,
        obj: Any,
        header: Optional[bytes] = None,
        topic: Optional[str] = None,
        *,
        is_key: Optional[bool] = None,
    ) -> bytes:
        return json.dumps(obj, default=str).encode()



import os
import json
from typing import Any, Optional

from wunderkafka.logger import logger
from wunderkafka.config.krb.rdkafka import init_kerberos
from wunderkafka.schema_registry.abc import AbstractHTTPClient
from wunderkafka.config.schema_registry import SRConfig
from wunderkafka.schema_registry.compat import http
from wunderkafka.config.krb.schema_registry import HTTPKerberosAuth


class KerberizableHTTPClient(AbstractHTTPClient):
    def __init__(
        self,
        config: SRConfig,
        *,
        requires_kerberos: bool = False,
        save_replay: bool = False,
        cmd_kinit: Optional[str] = None,
        krb_timeout: int = 60,
    ) -> None:
        s = http.Session()
        if requires_kerberos and cmd_kinit is not None:
            init_kerberos(cmd_kinit, krb_timeout)
            s.auth = HTTPKerberosAuth(
                principal=config.sasl_username,
                mutual_authentication=config.mutual_auth,
            )

        if config.ssl_certificate_location is not None:
            if config.ssl_key_location is not None:
                s.cert = (config.ssl_certificate_location, config.ssl_key_location)
            else:
                s.cert = config.ssl_certificate_location

        if config.ssl_key_location is not None:
            s.verify = config.ssl_ca_location

        accept = ", ".join(
            [
                "application/vnd.schemaregistry.v1+json",
                "application/vnd.schemaregistry+json",
                "application/json",
            ]
        )
        s.headers.update(
            {
                "Accept": accept,
                "Content-Type": "application/vnd.schemaregistry.v1+json",
            }
        )

        self._session = s
        self._base_url = config.url

        self._save_replay = save_replay

    @property
    def base_url(self) -> str:
        return str(self._base_url)

    def close(self) -> None:
        self._session.close()

    def make_request(
        self,
        relative_url: str,
        method: str = "GET",
        body: Any = None,
        query: Any = None,
    ) -> Any:
        url = "/".join([self._base_url.rstrip("/"), relative_url.lstrip("/")])
        logger.debug(f"{method}: {url}")
        response = self._session.request(method, url, json=body, params=query)
        # ToDo(aa.perelygin): more informative message when fail
        if response.status_code >= 400:
            logger.error(f"HTTPError for url: {url}, more details: {response.content}")

        response.raise_for_status()

        self._dump(method, relative_url, response)
        return response.json()

    def _dump(self, method: str, relative_url: str, response: http.Response) -> None:
        if self._save_replay:
            filename = f"{method}/{relative_url}.json"
            dir_name = os.path.dirname(filename)
            os.makedirs(dir_name, exist_ok=True)
            with open(filename, "w") as fl:
                json.dump(response.json(), fl)



from typing import Any, Union, Optional

from pydantic_settings import BaseSettings

from wunderkafka.config.krb.schema_registry import HTTPKerberosMutualAuth


def remap_fields(dct: dict[str, Any]) -> dict[str, Any]:
    return {f_name.replace("_", "."): f_value for f_name, f_value in dct.items()}


class SRConfig(BaseSettings):
    url: str
    ssl_ca_location: Optional[str] = None
    ssl_key_location: Optional[str] = None
    ssl_certificate_location: Optional[str] = None
    basic_auth_user_info: Optional[str] = None

    # ToDo (tribunsky-kir): looks like sasl_username and kerberos over HTTP are cloudera-specific.
    sasl_username: Optional[str] = None
    # ToDo: (tribunsky-kir): I'd prefer to compose the whole HTTP Kerberos stuff as separate subconfig,
    #                        but it entails writing additional logic for sasl username reuse.
    mutual_auth: Optional[HTTPKerberosMutualAuth] = None

    def dict(self, **kwargs: Any) -> dict[str, Optional[Union[str, int]]]:
        dct = super().model_dump(**kwargs)
        return remap_fields(dct)



"""
This module contains the high-level pipeline to produce messages with a nested consumer.

It is intended to be testable enough due to the composition of dependencies.

All moving parts should be interchangeable in terms of schema, header and serialization handling
(for further overriding^W extending).
"""

import datetime
from typing import Any, Union, TypeVar, Optional

from confluent_kafka import Message, TopicPartition
from confluent_kafka.serialization import MessageField, SerializationError, SerializationContext

from wunderkafka.types import HeaderParser
from wunderkafka.logger import logger
from wunderkafka.serdes.abc import AbstractDeserializer
from wunderkafka.structures import SchemaMeta, SerializerSchemaDescription
from wunderkafka.consumers.abc import AbstractConsumer, AbstractDeserializingConsumer
from wunderkafka.consumers.types import PayloadError, StreamResult
from wunderkafka.schema_registry.abc import AbstractSchemaRegistry
from wunderkafka.consumers.subscription import TopicSubscription

T = TypeVar("T")


def choose_one_of(
    common_deserializer: Optional[AbstractDeserializer],
    specific_deserializer: Optional[AbstractDeserializer],
    what: str,
) -> AbstractDeserializer:
    chosen_deserializer = specific_deserializer if specific_deserializer else common_deserializer
    if chosen_deserializer is None:
        msg = [
            f"{what.capitalize()} deserializer is not specified,",
            f"it should be passed via {what}_deserializer or deserializer at least.",
        ]
        raise ValueError(" ".join(msg))
    return chosen_deserializer


class HighLevelDeserializingConsumer(AbstractDeserializingConsumer):
    """Deserializing pipeline implementation of extended consumer."""

    def __init__(
        self,
        consumer: AbstractConsumer,
        headers_handler: Optional[HeaderParser],
        schema_registry: Optional[AbstractSchemaRegistry],
        deserializer: Optional[AbstractDeserializer] = None,
        value_deserializer: Optional[AbstractDeserializer] = None,
        key_deserializer: Optional[AbstractDeserializer] = None,
        *,
        stream_result: bool = False,
    ):
        """
        Init consumer with specific dependencies.

        :param consumer:            Consumer implementation to receive messages.
        :param headers_handler:     Callable to parse binary headers.
        :param schema_registry:     Schema registry client.
        :param deserializer:        Common message deserializer for value and key, if set.
        :param value_deserializer:  Message deserializer for value, if set.
        :param key_deserializer:    Message deserializer for the key, if set.
        :param stream_result:       If True, return a complex StreamResult object instead of just a model.
        """
        self.consumer = consumer
        self._header_parser = headers_handler
        self._registry = schema_registry
        self._deserializer = deserializer

        self._value_deserializer = choose_one_of(deserializer, value_deserializer, "value")
        self._key_deserializer = choose_one_of(deserializer, key_deserializer, "key")

        self._stream_result = stream_result

    def subscribe(  # noqa: D102,WPS211 # docstring inherited from superclass.
        self,
        topics: list[Union[str, TopicSubscription]],
        *,
        from_beginning: Optional[bool] = None,
        offset: Optional[int] = None,
        ts: Optional[int] = None,
        dt: Optional[datetime.datetime] = None,
        with_timedelta: Optional[datetime.timedelta] = None,
    ) -> None:
        self.consumer.subscribe(
            topics,
            from_beginning=from_beginning,
            offset=offset,
            ts=ts,
            with_timedelta=with_timedelta,
        )

    def commit(  # noqa: D102,WPS211 # docstring inherited from superclass.
        self,
        message: Optional[Message] = None,
        offsets: Optional[list[TopicPartition]] = None,
        asynchronous: bool = True,
    ) -> Optional[list[TopicPartition]]:
        if message is None and offsets is not None:
            return self.consumer.commit(offsets=offsets, asynchronous=asynchronous)
        if message is not None and offsets is None:
            return self.consumer.commit(message=message, asynchronous=asynchronous)
        # Default behavior
        return self.consumer.commit(message=message, offsets=offsets, asynchronous=asynchronous)

    def consume(
        self,
        timeout: float = 1.0,
        num_messages: int = 1000000,
        *,
        ignore_keys: bool = False,
        raise_on_error: bool = True,
        raise_on_lost: bool = True,
    ) -> list[T]:
        """
        Consume as many messages as we can for a given timeout and decode them.

        :param timeout:         The maximum time to block waiting for messages. Decoding time doesn't count.
        :param num_messages:    The maximum number of messages to receive from broker.
                                Default is 1000000 that was the allowed maximum for librdkafka 1.2.
        :param ignore_keys:     If True, skip key decoding, key will be set to None. Otherwise, decode key as usual.
        :param raise_on_error:  If True, raise KafkaError form confluent_kafka library to handle in client code.
        :param raise_on_lost:   If True, check on own clocks if max.poll.interval.ms is exceeded. If so, raises
                                ConsumerException to be handled in client code.

        :raises KafkaError:     See https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#confluent_kafka.KafkaError  # noqa: E501

        :return:                A list of Message objects with decoded value() and key() (possibly empty on timeout).
        """  # noqa: E501
        msgs = self.consumer.batch_poll(timeout, num_messages, raise_on_lost=raise_on_lost)
        return self._decoded(
            msgs,
            ignore_keys=ignore_keys,
            raise_on_error=raise_on_error,
        )

    def _decoded(self, msgs: list[Message], *, ignore_keys: bool, raise_on_error: bool) -> list[T]:
        results: list[StreamResult] = []
        for msg in msgs:
            kafka_error = msg.error()
            # Not sure if there is any need for an option to exclude errored message from the consumed ones,
            # so there is no `else`. In case when `.error()` returns KafkaError, but `raise_on_error` is set to False,
            # `.value()` called in client code will return raw bytes with string from KafkaError,
            # e.g.
            #   b'Application maximum poll interval (300000ms) exceeded by Nms'
            if kafka_error is not None:
                logger.error(kafka_error)
                if raise_on_error:
                    # Even PyCharm stubs show that it is inherited from an object, in fact it is a valid Exception
                    raise kafka_error

            topic = msg.topic()

            raw_key_value = msg.key()
            decode_key_ok = True
            if ignore_keys:
                # Yes, we lose information, but it is necessary to not get raw bytes
                # if `.key()` will be called in client code later.
                msg.set_key(None)
            else:
                try:
                    decoded_key = self._decode(topic, msg.key(), is_key=True)
                # KeyDeserializationError is inherited from SerializationError
                except SerializationError:
                    decode_key_ok = False
                    logger.error(f"Unable to decode key from bytes: {raw_key_value}")
                    if not ignore_keys:
                        raise
                else:
                    msg.set_key(decoded_key)

            try:
                decoded_value = self._decode(topic, msg.value())
            except (SerializationError, ValueError) as exc:
                logger.error(f"Unable to decode value from bytes: {msg.value()}")
                if not self._stream_result:
                    raise
                value_error = str(exc)
                if decode_key_ok:
                    error = PayloadError(description=value_error)
                else:
                    message = f"Unable to decode key (topic: {topic}, key payload: {raw_key_value})"
                    error = PayloadError(description=message)
                results.append(StreamResult(payload=None, error=error, msg=msg))
            else:
                msg.set_value(decoded_value)
                if self._stream_result:
                    results.append(StreamResult(payload=decoded_value, error=None, msg=msg))

        return results if self._stream_result else msgs  # type: ignore[return-value]

    # Todo (tribunsky.kir): arguable: make different composition (headers, SR & deserializer united cache)
    def _decode(self, topic: str, blob: Optional[bytes], *, is_key: bool = False) -> Any:
        if blob is None:
            return None

        deserializer = self._get_deserializer(is_key)
        if deserializer.schemaless:
            return deserializer.deserialize("", blob)
        # Header is separate in the sake of customization, e.g., we don't have SR and put schema directly in a message
        assert self._header_parser is not None
        assert self._registry is not None
        parsed_header = self._header_parser(blob)
        ctx = SerializationContext(topic, MessageField.VALUE)
        deserializer.set_context(ctx)
        schema_meta = SchemaMeta(
            topic=topic,
            is_key=is_key,
            header=parsed_header,
        )
        schema_text = self._registry.get_schema_text(schema_meta)
        schema = SerializerSchemaDescription(text=schema_text)
        # performance tradeoff: a message may be long, and we don't want to:
        # - copy the whole tail
        # - have implicit offset as if we read buffer when extracting header
        # so dealing with implicit offset and the whole binary string
        return deserializer.deserialize(schema.text, blob, seek_pos=parsed_header.size)

    def _get_deserializer(self, is_key: bool) -> AbstractDeserializer:
        return self._key_deserializer if is_key else self._value_deserializer



import json

from pydantic import BaseModel
from pydantic_core import CoreSchema
from pydantic.json_schema import DEFAULT_REF_TEMPLATE, JsonSchemaMode, JsonSchemaValue, GenerateJsonSchema

from wunderkafka.compat import ParamSpec

P = ParamSpec("P")


class JSONClosedModelGenerator(GenerateJsonSchema):
    def __init__(
        self,
        by_alias: bool = True,
        ref_template: str = DEFAULT_REF_TEMPLATE,
        *,
        evolvable: bool = True,
        # bug https://github.com/python/typing/discussions/1286 or not,
        # but in the new version of python should work something like **P w/o ParamSpec definition
        **kwargs: P.kwargs,  # type: ignore[valid-type]
    ) -> None:
        super().__init__(by_alias, ref_template, **kwargs)
        self.__evolvable = evolvable

    def generate(self, schema: CoreSchema, mode: JsonSchemaMode = "validation") -> JsonSchemaValue:
        json_schema = super().generate(schema, mode=mode)
        if self.__evolvable:
            key_additional_properties = "additionalProperties"
            key_definitions = "$defs"
            json_schema[key_additional_properties] = False
            definitions = json_schema.get(key_definitions, {})
            for definition in definitions:
                if definitions[definition].get("type") == "object":
                    json_schema[key_definitions][definition][key_additional_properties] = False
        return json_schema


def derive(model_type: type[BaseModel], schema_generator: type[GenerateJsonSchema]) -> str:
    return json.dumps(model_type.model_json_schema(schema_generator=schema_generator))



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Union, Optional

from confluent_kafka.serialization import SerializationContext

from wunderkafka.types import TopicName, KeySchemaDescription, ValueSchemaDescription
from wunderkafka.structures import SRMeta, ParsedHeader


class AbstractProtocolHandler(ABC):
    @abstractmethod
    def parse(self, blob: bytes) -> ParsedHeader: ...

    # ToDo (tribunsky.kir): make it symmetrical and more general
    @abstractmethod
    def pack(self, protocol_id: int, meta: SRMeta) -> bytes: ...


# ToDo (tribunsky.kir): make it parametrized generic?
class AbstractDeserializer(ABC):
    schemaless: bool = False
    ctx: Optional[SerializationContext] = None

    def set_context(self, ctx: SerializationContext) -> None:
        self.ctx = ctx

    @abstractmethod
    def deserialize(self, schema: str, payload: bytes, seek_pos: Optional[int] = None) -> Any: ...


class AbstractSerializer(ABC):
    # We allow to nest store in serialized to avoid the necessity of passing it as tuples,
    # cause in general we want to split schema storing and message serialization,
    # and it's easier to compose like this.
    # Therefore, we can use Producer's single store, which was the single one on the old API,
    # but if serializer has its own store, it will be used instead.
    # Moreover, as for every serializer store may be passed manually,
    # it may be the same object or different if there is such a need.
    store: Optional[AbstractDescriptionStore] = None

    @abstractmethod
    def serialize(
        self,
        schema: str,
        payload: Any,
        header: Optional[bytes] = None,
        topic: Optional[str] = None,
        *,
        is_key: Optional[bool] = None,
    ) -> bytes: ...


class AbstractDescriptionStore(ABC):
    def __init__(self) -> None:
        self._values: dict[TopicName, ValueSchemaDescription] = {}
        self._keys: dict[TopicName, KeySchemaDescription] = {}

    def get(
        self,
        topic: TopicName,
        *,
        is_key: bool = False,
    ) -> Optional[Union[ValueSchemaDescription, KeySchemaDescription]]:
        if is_key:
            return self._keys.get(topic)
        else:
            return self._values.get(topic)

    @abstractmethod
    def add(self, topic: TopicName, value: Any, key: Any) -> None: ...



from abc import ABC, abstractmethod
from typing import Any

from wunderkafka.structures import SRMeta, SchemaMeta, SchemaType


# TODO: replace with http.HTTPMethod when we drop python 3.10
class HTTPMethod:
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"


class AbstractHTTPClient(ABC):
    @abstractmethod
    def make_request(self, relative_url: str, method: str = "GET", body: Any = None, query: Any = None) -> Any: ...

    @property
    def base_url(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def close(self) -> None: ...

    def get(self, relative_url: str, body: Any = None, query: Any = None, **_: Any) -> Any:
        return self.make_request(relative_url, HTTPMethod.GET, body, query)

    def post(self, relative_url: str, body: Any = None, query: Any = None, **_: Any) -> Any:
        return self.make_request(relative_url, HTTPMethod.POST, body, query)

    def put(self, relative_url: str, body: Any = None, query: Any = None, **_: Any) -> Any:
        return self.make_request(relative_url, HTTPMethod.PUT, body, query)

    def patch(self, relative_url: str, body: Any = None, query: Any = None, **_: Any) -> Any:
        return self.make_request(relative_url, HTTPMethod.PATCH, body, query)

    def delete(self, relative_url: str, body: Any = None, query: Any = None, **_: Any) -> Any:
        return self.make_request(relative_url, HTTPMethod.DELETE, body, query)


class AbstractSchemaRegistry(ABC):
    @abstractmethod
    def get_schema_text(self, schema_meta: SchemaMeta) -> str: ...

    @abstractmethod
    def register_schema(
        self, topic: str, schema_text: str, schema_type: SchemaType, *, is_key: bool = True
    ) -> SRMeta: ...



from typing import Any


class SimpleCache:
    def __init__(self) -> None:
        self._store: dict[Any, Any] = {}

    def get(self, key: Any) -> Any:
        return self._store.get(key)

    def set(self, key: Any, value: Any) -> Any:
        self._store[key] = value


class AlwaysEmptyCache(SimpleCache):
    def set(self, key: Any, value: Any) -> Any:
        return None



from __future__ import annotations

from typing import Any, Optional

from typing_extensions import Protocol
from confluent_kafka.schema_registry import Schema
from confluent_kafka.schema_registry import SchemaRegistryClient as ConfluentSchemaRegistryClient

from wunderkafka.compat import ParamSpec
from wunderkafka.structures import SRMeta, SchemaMeta, SchemaType
from wunderkafka.schema_registry import SimpleCache
from wunderkafka.schema_registry.abc import AbstractHTTPClient, AbstractSchemaRegistry

P = ParamSpec("P")


class ConfluentRestClient(Protocol):
    def get(self, url: str, query: Optional[dict] = None) -> Any: ...

    def post(self, url: str, body: Optional[str], **kwargs: Any) -> Any: ...

    def delete(self, url: str) -> Any: ...

    def put(self, url: str, body: Optional[str] = None) -> Any: ...

    def send_request(self, url: str, method: str, body: Optional[str] = None, query: Optional[dict] = None) -> Any: ...

    def _close(self) -> None: ...


class SchemaRegistryHttpClientAdapter:
    """
    Adapter for AbstractHTTPClient and confluent kafka's _RestClient. Though
    AbstractHTTPClient supports all methods signatures are a bit different.
    """

    def __init__(self, http_client: AbstractHTTPClient) -> None:
        self._client = http_client

    def get(self, url: str, query: Optional[dict] = None) -> Any:
        return self._client.get(url, query=query)

    def post(self, url: str, body: Optional[str], **_: Any) -> Any:
        return self._client.post(url, body=body)

    def delete(self, url: str) -> Any:
        return self._client.delete(url)

    def put(self, url: str, body: Optional[str] = None) -> Any:
        return self._client.put(url, body=body)

    def send_request(self, url: str, method: str, body: Optional[str] = None, query: Optional[dict] = None) -> Any:
        return self._client.make_request(url, method, body=body, query=query)

    def _close(self) -> None:
        return self._client.close()


class SchemaRegistryClient(ConfluentSchemaRegistryClient):
    @classmethod
    def from_client(cls, http_client: AbstractHTTPClient, *args: P.args, **kwargs: P.kwargs) -> SchemaRegistryClient:  # type: ignore[valid-type]
        # Minimal initialization as we will override a client with our own
        client = cls({"url": http_client.base_url, **kwargs})
        client._rest_client = SchemaRegistryHttpClientAdapter(http_client)
        return client


class ConfluentSRClient(AbstractSchemaRegistry):
    def __init__(
        self,
        http_client: AbstractHTTPClient,
        _: Optional[SimpleCache] = None,
        *args: P.args,  # type: ignore[valid-type]
        **kwargs: P.kwargs,  # type: ignore[valid-type]
    ) -> None:
        self.client = SchemaRegistryClient.from_client(http_client, *args, **kwargs)

    def get_schema_text(self, meta: SchemaMeta) -> str:
        schema = self.client.get_schema(meta.header.schema_id)
        return schema.schema_str

    def register_schema(self, topic: str, schema_text: str, schema_type: SchemaType, *, is_key: bool = True) -> SRMeta:
        # FixMe (tribunsky.kir): lack of symmetry here - SchemaMeta knows about different vendors, but not vice versa.
        subject = "{}-{}".format(topic, "key" if is_key else "value")
        schema_id = self.client.register_schema(subject, Schema(schema_text, schema_type=schema_type))
        return SRMeta(schema_id, schema_version=None)



import os
from enum import Enum
from types import MappingProxyType
from typing import Any, Union, TypeVar, Optional
from collections.abc import Mapping

from wunderkafka.logger import logger
from wunderkafka.config.schema_registry import SRConfig
from wunderkafka.config.generated.fields import COMMON_FIELDS, CONSUMER_FIELDS, PRODUCER_FIELDS
from wunderkafka.config.generated.models import RDConsumerConfig, RDProducerConfig

CONF_CONSUMER_FIELDS = MappingProxyType(
    {field_name.replace(".", "_"): field_name for field_name in (*COMMON_FIELDS, *CONSUMER_FIELDS)}
)

CONF_PRODUCER_FIELDS = MappingProxyType(
    {field_name.replace(".", "_"): field_name for field_name in (*COMMON_FIELDS, *PRODUCER_FIELDS)}
)

ConfigValues = Union[str, int, bool, float]


def remap_properties(
    dct: dict[str, Optional[ConfigValues]],
    mapping: Mapping[str, str],
) -> dict[str, ConfigValues]:
    new_dct = {}
    for f_name, f_value in dct.items():
        if f_value is not None:
            to_add = f_value
            if isinstance(f_value, Enum):
                to_add = f_value.value
            new_dct[mapping[f_name]] = to_add
    return new_dct


# ToDo (tribunsky.kir): I need separate common place to do somthing with config before feeding it to librdkafka
#                       or write more complex generator and class hierarchy (e.g. platform-specific base classes)
#                       to not monkeypatch config before actually feeding it to librdkafka.
#                       #TypeSafety!!1
def sanitize(dct: dict[str, ConfigValues]) -> dict[str, ConfigValues]:
    if os.name == "nt":
        exclude = {
            # cimpl.KafkaException: KafkaError{
            #   ...
            #   "Configuration property "sasl.kerberos.kinit.cmd" not supported in this build: Kerberos keytabs are not supported on Windows, instead the logged on user's credentials are used through native SSPI"
            #  }
            "sasl.kerberos.kinit.cmd": "Kerberos keytabs are not supported on Windows, instead the logged on user's credentials are used through native SSPI"
        }
    else:
        exclude = {
            # cimpl.KafkaException: KafkaError{
            #   ...
            #   "Configuration property "ssl.ca.certificate.stores" not supported in this build: configuration only valid on Windows"
            #  }
            "ssl.ca.certificate.stores": 'Configuration property "ssl.ca.certificate.stores" not supported in this build: configuration only valid on Windows'
        }
    for property_name, exclude_reason in exclude.items():
        property_value = dct.pop(property_name, None)
        if property_value is not None:
            logger.warning(f"Excluding {property_name}={property_value} ({exclude_reason}) even it was set to default")
    return dct


# I don't like mixing SR config and librdkafka config,
# but it's more handful for producer w/o schema (no need to nest config for librdkafka)
class ConsumerConfig(RDConsumerConfig):
    # https://docs.pydantic.dev/latest/migration/#required-optional-and-nullable-fields
    sr: Optional[SRConfig] = None

    def dict(self, **kwargs: Any) -> dict[str, ConfigValues]:
        dct = super().model_dump(**kwargs)
        dct.pop("sr")
        return sanitize(remap_properties(dct, CONF_CONSUMER_FIELDS))


class ProducerConfig(RDProducerConfig):
    # https://docs.pydantic.dev/latest/migration/#required-optional-and-nullable-fields
    sr: Optional[SRConfig] = None

    def dict(self, **kwargs: Any) -> dict[str, ConfigValues]:
        dct = super().model_dump(**kwargs)
        dct.pop("sr")
        return sanitize(remap_properties(dct, CONF_PRODUCER_FIELDS))


RDKafkaConfig = TypeVar("RDKafkaConfig", ConsumerConfig, ProducerConfig)



"""This module contains some predefined callbacks to interact with librdkafka."""

from __future__ import annotations

from confluent_kafka import Message, KafkaError

from wunderkafka.logger import logger


def info_callback(err: KafkaError | None, msg: Message) -> None:
    """
    Log every message delivery.

    :param err:             Error, if any, thrown from confluent-kafka cimpl.
    :param msg:             Message to be delivered.
    """
    if err is None:
        logger.info(f"Message delivered to {msg.topic()} partition: {msg.partition()}")
    else:
        logger.error(f"Message failed delivery: {err}")


def error_callback(err: KafkaError | None, _: Message) -> None:
    """
    Log only failed message delivery.

    :param err:             Error, if any, thrown from confluent-kafka cimpl.
    :param _:               Message to be delivered (unused, but needed to not break callback signature).
    """
    if err:
        logger.error(f"Message failed delivery: {err}")



"""This module contains implementation of extended confluent-kafka Producer's API."""

from __future__ import annotations

import atexit

from confluent_kafka import KafkaException

from wunderkafka.types import DeliveryCallback
from wunderkafka.compat import ParamSpec
from wunderkafka.config import ProducerConfig
from wunderkafka.producers.abc import AbstractProducer
from wunderkafka.callbacks.producer import error_callback
from wunderkafka.config.krb.rdkafka import challenge_krb_arg

P = ParamSpec("P")


class BytesProducer(AbstractProducer):
    """Producer implementation of extended interface for raw messages."""

    def __init__(self, config: ProducerConfig) -> None:
        """
        Init producer.

        :param config:          Pydantic model with librdkafka producer's configuration.
        """
        try:
            super().__init__(config.dict())
        except KafkaException as exc:
            config = challenge_krb_arg(exc, config)
            super().__init__(config.dict())

        self._config = config
        # TODO (tribunsky.kir): make it configurable
        #                       https://github.com/severstal-digital/wunderkafka/issues/90
        atexit.register(self.flush)

    # TODO (tribunsky.kir): Do we need re-initiation of consumer/producer in runtime?
    #                       https://github.com/severstal-digital/wunderkafka/issues/94
    @property
    def config(self) -> ProducerConfig:
        """
        Get the producer's config.

        :return:        Pydantic model with librdkafka producer's configuration.
        """
        return self._config

    # TODO (tribunsky.kir): rethink API?
    #                       https://github.com/severstal-digital/wunderkafka/issues/91
    # https://github.com/python/mypy/issues/13966
    def send_message(  # type: ignore[valid-type] # noqa: PLR0913
        self,
        topic: str,
        value: str | bytes | None = None,
        key: str | bytes | None = None,
        partition: int | None = None,
        on_delivery: DeliveryCallback | None = error_callback,
        *args: P.args,  # type: ignore[valid-type] # noqa: ARG002
        blocking: bool = False,
        **kwargs: P.kwargs,  # type: ignore[valid-type]
    ) -> None:
        if partition is not None:
            self.produce(topic, value, key=key, partition=partition, on_delivery=on_delivery, **kwargs)
        else:
            self.produce(topic, value, key=key, on_delivery=on_delivery, **kwargs)
        if blocking:
            self.flush()
        else:
            self.poll(0)



from __future__ import annotations

import datetime

from wunderkafka.message import MessageProtocol as Message
from wunderkafka.consumers.bytes import BytesConsumer
from wunderkafka.consumers.subscription import TopicSubscription


class TestConsumer(BytesConsumer):
    def __init__(self, msgs: list[Message]) -> None:
        self._msgs = msgs

    def subscribe(
        self,
        topics: list[str | TopicSubscription],
        *,
        from_beginning: bool | None = None,
        offset: int | None = None,
        ts: float | None = None,
        dt: datetime.datetime | None = None,
        with_timedelta: datetime.timedelta | None = None,
    ) -> None:
        pass

    def batch_poll(
        self,
        timeout: float = 1.0,
        num_messages: int = 1000000,
        *,
        raise_on_lost: bool = False,
    ) -> list[Message]:
        return self._msgs



from __future__ import annotations

import time
from typing import Union, Generic, TypeVar, Optional

from pydantic import Field, BaseModel, ConfigDict, model_validator
from confluent_kafka import Message

from wunderkafka.message import MessageProtocol

T = TypeVar("T")


class PayloadError(BaseModel):
    description: str


class StreamResult(BaseModel, Generic[T]):
    # it's needed to allow attaching the real cimpl.Message in runtime anyway
    model_config = ConfigDict(arbitrary_types_allowed=True)

    payload: Optional[T] = None
    error: Optional[PayloadError] = None
    # not via `|`, because pydantic raises TypeError in python 3.9 without `eval_type_backport` package despite
    # `from __future__ import annotations`
    msg: Union[Message, MessageProtocol]
    t0: float = Field(default_factory=time.perf_counter)

    @property
    def ok(self) -> bool:
        return self.error is None and self.payload is not None

    @property
    def lifetime(self) -> float:
        return time.perf_counter() - self.t0

    @model_validator(mode="after")
    def verify_mutually_exclusive(self) -> StreamResult:
        # The payload may be None due to Kafka tombstones:
        # https://docs.confluent.io/kafka/design/log_compaction.html
        # In short, these are messages with a null value, used to indicate that their key has been deleted.
        if self.payload is not None and self.error is not None:
            raise ValueError("payload and error cannot both be not None")
        return self



from typing import Any, Optional

from wunderkafka import BytesConsumer, BytesProducer, ConsumerConfig, ProducerConfig
from wunderkafka.types import TopicName, MessageDescription
from wunderkafka.consumers.constructor import HighLevelDeserializingConsumer
from wunderkafka.producers.constructor import HighLevelSerializingProducer
from wunderkafka.serdes.schemaless.json.serializers import SchemaLessJSONSerializer
from wunderkafka.serdes.schemaless.json.deserializers import SchemaLessJSONDeserializer
from wunderkafka.serdes.schemaless.string.serializers import StringSerializer
from wunderkafka.serdes.schemaless.string.deserializers import StringDeserializer
from wunderkafka.serdes.schemaless.jsonmodel.serializers import SchemaLessJSONModelSerializer


class SchemaLessJSONStringProducer(HighLevelSerializingProducer):
    """Kafka Producer client to serialize and send any value as JSON without any schema."""

    def __init__(
        self,
        mapping: Optional[dict[TopicName, MessageDescription]],
        config: ProducerConfig,
    ) -> None:
        """
        Init producer from pre-defined blocks.

        :param mapping:     Topic-to-Schemas mapping.
                            Mapping's value should contain at least message's value schema to be used for serialization.
        :param config:      Configuration for:

                                - Librdkafka producer.
                                - Schema registry client (conventional options for HTTP).

                            Refer original CONFIGURATION.md (https://git.io/JmgCl) or generated config.
        """
        super().__init__(
            producer=BytesProducer(config),
            schema_registry=None,
            header_packer=None,
            value_serializer=SchemaLessJSONSerializer(),
            key_serializer=StringSerializer(),
            mapping=mapping,
        )


class SchemaLessJSONModelStringProducer(HighLevelSerializingProducer):
    """Kafka Producer client to serialize and send any instance of pydantic model as JSON without any schema."""

    def __init__(
        self,
        mapping: Optional[dict[TopicName, MessageDescription]],
        config: ProducerConfig,
        **serializer_kwargs: Any,
    ) -> None:
        """
        Init producer from pre-defined blocks.

        :param mapping:             Topic-to-Schemas mapping.
                                    Mapping's value should contain at least message's value schema to be used for serialization.
        :param config:              Configuration for:

                                        - Librdkafka producer.
                                        - Schema registry client (conventional options for HTTP).

                                    Refer original CONFIGURATION.md (https://git.io/JmgCl) or generated config.

        :param serializer_kwargs:   Additional keyword arguments to pass to the SchemaLessJSONModelSerializer,
                                    see https://docs.pydantic.dev/latest/api/base_model/#pydantic.BaseModel.model_dump_json
        """

        super().__init__(
            producer=BytesProducer(config),
            schema_registry=None,
            header_packer=None,
            value_serializer=SchemaLessJSONModelSerializer(**serializer_kwargs),
            key_serializer=StringSerializer(),
            mapping=mapping,
        )


class SchemaLessJSONStringConsumer(HighLevelDeserializingConsumer):
    """Kafka Consumer client to get JSON-serialized messages without any schema."""

    def __init__(self, config: ConsumerConfig) -> None:
        """
        Init consumer from pre-defined blocks.

        :param config:      Configuration for:

                                - Librdkafka consumer.
                                - Schema registry client (conventional options for HTTP).

                            Refer original CONFIGURATION.md (https://git.io/JmgCl) or generated config.
        """

        self._default_timeout: int = 60

        super().__init__(
            consumer=BytesConsumer(config),
            schema_registry=None,
            headers_handler=None,
            value_deserializer=SchemaLessJSONDeserializer(),
            key_deserializer=StringDeserializer(),
            stream_result=False,
        )



import json
from typing import Any, Optional

from confluent_kafka.schema_registry.json_schema import JSONDeserializer as JSONDSchemaDeserializer

from wunderkafka.serdes.abc import AbstractDeserializer


class JSONDeserializer(AbstractDeserializer):
    def __init__(self) -> None:
        self._cache: dict[str, JSONDSchemaDeserializer] = {}

    def deserialize(self, schema: str, blob: bytes, seek_pos: Optional[int] = None) -> Any:
        if schema not in self._cache:
            self._cache[schema] = JSONDSchemaDeserializer(json.dumps(json.loads(schema)))
        deserializer = self._cache[schema]
        dct = deserializer(blob, self.ctx)
        return dct



from typing import Any, Union, Optional

from wunderkafka.errors import SchemaRegistryLookupException
from wunderkafka.logger import logger
from wunderkafka.structures import SRMeta, SchemaMeta, SchemaType, ParsedHeader
from wunderkafka.serdes.vendors import Actions, get_protocol
from wunderkafka.schema_registry.abc import AbstractHTTPClient, AbstractSchemaRegistry
from wunderkafka.schema_registry.cache import SimpleCache, AlwaysEmptyCache


def prepare_schemas_for_caching(
    versions: dict[str, list[dict[str, Optional[Union[int, str]]]]],
) -> dict[ParsedHeader, str]:
    schemas_to_header_mapping = {}
    for schema in versions["entities"]:
        txt = schema["schemaText"]
        if not isinstance(txt, str):
            raise ValueError(f"Provided SchemaText is not str: {schema}")
        meta_id = schema["schemaMetadataId"]
        if not isinstance(meta_id, int):
            raise ValueError(f"Wrong schemaMetadataId (not int): {schema}")
        schema_version = schema["version"]
        if not isinstance(schema_version, int):
            raise ValueError(f"Wrong schema version (not int): {schema}")
        new_header_v1 = ParsedHeader(
            protocol_id=1,
            meta_id=meta_id,
            schema_version=schema_version,
            schema_id=None,
            size=get_protocol(1, Actions.deserialize).header_size + 1,
        )
        schema_id = schema["id"]
        if not isinstance(schema_id, int):
            raise ValueError(f"Wrong schema id (not int): {schema}")
        new_header_v2 = ParsedHeader(
            protocol_id=2,
            meta_id=None,
            schema_version=None,
            schema_id=schema_id,
            size=get_protocol(2, Actions.deserialize).header_size + 1,
        )
        new_header_v3 = ParsedHeader(
            protocol_id=3,
            meta_id=None,
            schema_version=None,
            schema_id=schema_id,
            size=get_protocol(3, Actions.deserialize).header_size + 1,
        )
        for new_header in (new_header_v1, new_header_v2, new_header_v3):
            schemas_to_header_mapping[new_header] = txt
    return schemas_to_header_mapping


class ClouderaSRClient(AbstractSchemaRegistry):
    def __init__(self, http_client: AbstractHTTPClient, cache: Optional[SimpleCache] = None) -> None:
        self._client = http_client
        self._cache = AlwaysEmptyCache() if cache is None else cache
        self._requests_count = 0

    @property
    def requests_count(self) -> int:
        return self._requests_count

    def _send_request_for_schema(self, meta: SchemaMeta) -> Any:
        versions = self._client.get(f"schemas/{meta.subject}/versions")
        self._requests_count += 1
        return versions

    def _create_meta(self, subject: str) -> int:
        body = {
            "type": "avro",
            "schemaGroup": "Kafka",
            "name": subject,
            "description": subject,
            "compatibility": "BACKWARD",
            "validationLevel": "ALL",
        }
        return self._client.post("schemas", body=body)

    def _create_schema(self, subject: str, schema_text: str) -> int:
        body = {
            "description": subject,
            # TypeError: Object of type 'mappingproxy' is not JSON serializable
            # https://github.com/confluentinc/confluent-kafka-python/issues/610
            "schemaText": schema_text,
        }
        query = {"branch": "MASTER"}
        return self._client.post(f"schemas/{subject}/versions", body=body, query=query)

    def get_schema_text(self, meta: SchemaMeta) -> str:
        # TODO (tribunsky.kir): arguably, the best key is BINARY header, not dataclass.
        # TODO (tribunsky.kir): ineffective caching. When we get response with list of schemas,
        #                       we can do better: pre-generate all possible metas and keep texts
        schema_text = self._cache.get(meta.header)
        if schema_text is not None:
            return schema_text
        logger.debug(f"Couldn't find header ({meta.header}) in cache, re-requesting...")
        versions = self._send_request_for_schema(meta)
        for hdr, txt in prepare_schemas_for_caching(versions).items():
            self._cache.set(hdr, txt)
        schema_text = self._cache.get(meta.header)
        if schema_text is None:
            error_message = f"Couldn't find schema for {meta.header}"
            logger.error(error_message)
            for schema in versions["entities"]:
                logger.warning(f"\tschema: {schema}")
            raise SchemaRegistryLookupException(error_message)
        return schema_text

    # TODO (tribunsky.kir): here and in producers/constructor:
    #                       - make hypothetical meta + prop for subject
    #                       - symmetry with get_schema_text
    def register_schema(self, topic: str, schema_text: str, schema_type: SchemaType, *, is_key: bool = True) -> SRMeta:
        uid = (topic, schema_text, is_key)
        meta = self._cache.get(uid)
        if meta is None:
            suffix = ":k" if is_key else ""
            subject = f"{topic}{suffix}"
            # TODO (tribunsky.kir): three queries is bad. Either we should know protocol id,
            #                       e.g. to skip last query (we do not need id always)
            meta_id = self._create_meta(subject)
            schema_version = self._create_schema(subject, schema_text)
            versions = self._client.get(f"schemas/{subject}/versions")
            for mt in versions["entities"]:
                if mt["schemaMetadataId"] == meta_id and mt["version"] == schema_version:
                    meta = SRMeta(
                        meta_id=meta_id,
                        schema_version=schema_version,
                        schema_id=mt["id"],
                    )
            self._cache.set(uid, meta)
        return meta



from typing import Any, Optional

from confluent_kafka.serialization import StringDeserializer as _StringDeserializer

from wunderkafka.serdes.abc import AbstractDeserializer


class StringDeserializer(AbstractDeserializer):
    def __init__(self) -> None:
        self.schemaless = True
        self._deserializer = _StringDeserializer()

    def deserialize(self, schema: str, blob: bytes, seek_pos: Optional[int] = None) -> Any:
        return self._deserializer(blob, None)



import json
from typing import Any, Optional

from pydantic import BaseModel
from confluent_kafka.serialization import MessageField, SerializationContext
from confluent_kafka.schema_registry import SchemaRegistryClient
from confluent_kafka.schema_registry.json_schema import JSONSerializer

from wunderkafka.serdes.abc import AbstractSerializer, AbstractDescriptionStore
from wunderkafka.serdes.store import JSONModelRepo


def pydantic_to_dict(model: BaseModel, _: SerializationContext) -> dict:
    dump = model.model_dump_json()
    dct = json.loads(dump)
    return dct


class JSONModelSerializer(AbstractSerializer):
    def __init__(
        self,
        schema_registry_client: SchemaRegistryClient,
        store: Optional[AbstractDescriptionStore] = None,
    ) -> None:
        self._cache: dict[str, JSONSerializer] = {}
        self._sr_client = schema_registry_client
        self.store = store or JSONModelRepo()

    def serialize(
        self,
        schema: str,
        obj: Any,
        header: Optional[bytes] = None,
        topic: Optional[str] = None,
        *,
        is_key: Optional[bool] = None,
    ) -> bytes:
        if schema not in self._cache:
            self._cache[schema] = JSONSerializer(schema, self._sr_client, pydantic_to_dict)
        serializer = self._cache[schema]
        field = MessageField.KEY if is_key else MessageField.VALUE
        return serializer(obj, SerializationContext(topic, field))



"""This module contains implementation of extended confluent-kafka Consumer's API."""

from __future__ import annotations

import time
import atexit
import datetime

from confluent_kafka import KafkaException

from wunderkafka.types import HowToSubscribe
from wunderkafka.config import ConsumerConfig
from wunderkafka.errors import ConsumerException
from wunderkafka.logger import logger
from wunderkafka.consumers.abc import Message, AbstractConsumer
from wunderkafka.config.krb.rdkafka import challenge_krb_arg
from wunderkafka.consumers.subscription import TopicSubscription
from wunderkafka.callbacks.consumer.subscribe import reset_partitions


class BytesConsumer(AbstractConsumer):
    """Consumer implementation of extended interface for raw messages."""

    def __init__(self, config: ConsumerConfig) -> None:
        """
        Init consumer.

        :param config:          Pydantic BaseSettings model with librdkafka consumer's configuration.
        """
        try:
            super().__init__(config.dict())
        except KafkaException as exc:
            config = challenge_krb_arg(exc, config)
            super().__init__(config.dict())
        self.subscription_offsets: dict[str, HowToSubscribe] | None = None

        self._config = config
        self._last_poll_ts = time.perf_counter()
        # TODO (tribunsky-kir): make it configurable
        #                       https://github.com/severstal-digital/wunderkafka/issues/90
        atexit.register(self.close)

    def __str__(self) -> str:
        """
        Get human-readable representation of consumer.

        :return:    string with consumer gid.
        """
        return f"{self.__class__.__name__}:{self._config.group_id}"

    def batch_poll(
        self,
        timeout: float = 1.0,
        num_messages: int = 1000000,
        *,
        raise_on_lost: bool = False,
    ) -> list[Message]:
        # TODO (tribunsky.kir): maybe it better to use on_lost callback within subscribe()
        #                       https://github.com/severstal-digital/wunderkafka/issues/95
        dt = int((time.perf_counter() - self._last_poll_ts) * 1000)
        if dt > self._config.max_poll_interval_ms:
            msg = f"Exceeded max.poll.interval.ms ({self._config.max_poll_interval_ms}): {dt}"

            if raise_on_lost:
                # TODO (tribunsky.kir): resubscribe by ts?
                #                       https://github.com/severstal-digital/wunderkafka/issues/95
                raise ConsumerException(msg)
            logger.warning(msg)

        msgs = self.consume(num_messages=num_messages, timeout=timeout)
        self._last_poll_ts = time.perf_counter()
        return msgs

    # TODO (tribunsky.kir): do not override original API and wrap it in superclass
    #                       https://github.com/severstal-digital/wunderkafka/issues/93
    def subscribe(  # noqa: PLR0913
        self,
        topics: list[str | TopicSubscription],
        *,
        from_beginning: bool | None = None,
        offset: int | None = None,
        ts: float | None = None,
        dt: datetime.datetime | None = None,
        with_timedelta: datetime.timedelta | None = None,
    ) -> None:
        """
        Subscribe to a list of topics, or a list of specific TopicSubscriptions.

        This method overrides original `subscribe()` method of `confluent-kafka.Consumer` and allows to subscribe
        to a topic via specific offset or timestamp.

        Doc of an original method depends on python's :code:`confluent-kafka` version. Please, refer to
        `confluent-kafka documentation <https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#pythonclient-consumer>`__.

        .. warning::
            Currently this method doesn't allow passing callbacks and uses its own to reset partitions.
        """
        subscriptions = {}
        for tpc in topics:
            if isinstance(tpc, str):
                tpcs = TopicSubscription(
                    topic=tpc,
                    from_beginning=from_beginning,
                    offset=offset,
                    ts=ts,
                    dt=dt,
                    with_timedelta=with_timedelta,
                )
            else:
                tpcs = tpc
            subscriptions[tpcs.topic] = tpcs.how

        # We have a specific subscription at least once
        if any(subscriptions.values()):
            self.subscription_offsets = subscriptions
            # TODO (tribunsky.kir): avoid mutation of self.subscription_offset and remove it as a field
            #                       https://github.com/severstal-digital/wunderkafka/issues/87
            super().subscribe(topics=list(self.subscription_offsets), on_assign=reset_partitions)
        else:
            super().subscribe(topics=list(subscriptions))



"""
This module contains the high-level pipeline to produce messages with a nested producer.

It is intended to be testable enough due to the composition of dependencies.

All moving parts should be interchangeable in terms of schema, header and serialization handling
(for further overriding^W extending).
"""

from typing import Any, Optional

from wunderkafka.types import MsgKey, MsgValue, TopicName, HeaderPacker, DeliveryCallback, MessageDescription
from wunderkafka.logger import logger
from wunderkafka.serdes.abc import AbstractSerializer, AbstractDescriptionStore
from wunderkafka.structures import SRMeta, DeserializerSchemaDescription
from wunderkafka.producers.abc import AbstractProducer, AbstractSerializingProducer
from wunderkafka.callbacks.producer import error_callback
from wunderkafka.schema_registry.abc import AbstractSchemaRegistry


class HighLevelSerializingProducer(AbstractSerializingProducer):
    """Serializing pipeline implementation of extended producer."""

    def __init__(  # noqa: WPS211  # ToDo (tribunsky.kir): rethink building of producer.
        #                                                  Maybe single Callable (like in kafka-python) much better.
        self,
        producer: AbstractProducer,
        # Some serializers doesn't need SR at all, e.g. StringSerializer.
        schema_registry: Optional[AbstractSchemaRegistry],
        # As some serializers doesn't contain magic byte, we do not need to handle the first bytes of a message at all.
        header_packer: Optional[HeaderPacker],
        serializer: Optional[AbstractSerializer] = None,
        store: Optional[AbstractDescriptionStore] = None,
        # ToDo: switch mapping to something like consumer's TopicSubscription?
        mapping: Optional[dict[TopicName, MessageDescription]] = None,
        value_serializer: Optional[AbstractSerializer] = None,
        key_serializer: Optional[AbstractSerializer] = None,
        *,
        protocol_id: int = 1,
        lazy: bool = False,
    ) -> None:
        """
        Init producer with specific dependencies and prepare it to work against specified topic(s).

        :param producer:            Producer implementation to send messages.
        :param schema_registry:     Schema registry client.
        :param header_packer:       Callable to form binary headers.
        :param serializer:          Common message serializer for the key and value.
                                    If specific value_deserializer/key_deserializer defined, it will be used instead.
        :param store:               Specific store to provide schema text extraction from schema description.
        :param mapping:             Per-topic definition of value and/or key schema descriptions.
        :param value_serializer:    Message serializer for value, if set.
        :param key_serializer:      Message serializer for the key, if set.
        :param protocol_id:         Protocol id for producer (1 - Cloudera, 0 - Confluent, etc.)
        :param lazy:                If True,
                                    defer schema registry publication, otherwise schema will be registered
                                    before the first message sending.
        """
        self._mapping = mapping or {}
        self._registered: set[TopicName] = set()

        # ToDo (tribunsky.kir): look like wrong place. Maybe it's better to highlight an entity
        #                       of Schema which may be checked or not. Then input mapping may be eliminated (or not).
        self._checked: dict[tuple[str, str, bool], SRMeta] = {}

        self._store = store
        self._sr = schema_registry
        self._serializer = serializer
        self._producer = producer
        self._header_packer = header_packer
        self._protocol_id = protocol_id

        chosen_value_serializer = value_serializer if value_serializer else serializer
        if chosen_value_serializer is None:
            msg = "Value serializer is not specified, should be passed via value_serializer or serializer at least."
            raise ValueError(msg)
        self._value_serializer = chosen_value_serializer

        chosen_key_serializer = key_serializer if value_serializer else serializer
        if chosen_key_serializer is None:
            msg = "Key serializer is not specified, should be passed via key_serializer or serializer at least."
            raise ValueError(msg)
        self._key_serializer = chosen_key_serializer

        for topic, description in self._mapping.items():
            if isinstance(description, (tuple, list)):
                msg_value, msg_key = description
                self.set_target_topic(topic, msg_value, msg_key, lazy=lazy)
            else:
                self.set_target_topic(topic, description, lazy=lazy)

    def flush(self, timeout: Optional[float] = None) -> int:  # noqa: D102 # docstring inherited from superclass.
        if timeout is None:
            return self._producer.flush()
        return self._producer.flush(timeout)

    def set_target_topic(  # noqa: D102 # docstring inherited from superclass.
        self,
        topic: str,
        value: Any,  # noqa: WPS110 # Domain. inherited from superclass.
        key: Any = None,
        *,
        lazy: bool = False,
    ) -> None:
        value_store = self._get_store(self._value_serializer)
        key_store = self._get_store(self._key_serializer)
        value_store.add(topic, value, None)
        if key is not None:
            # FixMe (tribunsky.kir): make key and value stores independent,
            #                        because now we cannot put None instead of value,
            #                        even though we need store only for the key.
            key_store.add(topic, value, key)
        if not lazy:
            value_descr = value_store.get(topic)
            self._check_schema(topic, value_descr)
            if key is not None:
                key_descr = key_store.get(topic, is_key=True)
                assert key_descr is not None
                if not key_descr.empty:
                    self._check_schema(topic, key_descr, is_key=True)
        self._registered.add(topic)

    def send_message(  # noqa: D102,WPS211 # inherited from superclass.
        self,
        topic: str,
        value: MsgValue = None,  # noqa: WPS110 # Domain. inherited from superclass.
        key: MsgKey = None,
        partition: Optional[int] = None,
        on_delivery: Optional[DeliveryCallback] = error_callback,
        *args: Any,
        blocking: bool = False,
        protocol_id: Optional[int] = None,
        **kwargs: Any,
    ) -> None:
        if topic not in self._registered:
            raise ValueError(f"Topic {topic} not registered while producing the message.")
        if protocol_id is None:
            protocol_id = self._protocol_id
        encoded_value = self._encode(topic, value, protocol_id)
        encoded_key = self._encode(topic, key, protocol_id, is_key=True)

        self._producer.send_message(
            topic,
            encoded_value,
            encoded_key,
            partition,
            on_delivery,
            blocking=blocking,
            *args,
            **kwargs,
        )

    def _check_schema(
        self,
        topic: str,
        schema: Optional[DeserializerSchemaDescription],
        *,
        is_key: bool = False,
        force: bool = False,
    ) -> Optional[SRMeta]:
        """
        Ensure that we have schema's ids necessary to create message's header.

        :param topic:   Target topic against which we are working for this call.
        :param schema:  Schema container to be registered. Should contain text.
        :param is_key:  If True, schema will be requested as a message key, message value otherwise.
        :param force:   If True, do not reuse cached results, always request Schema Registry.

        :raises ValueError: If received no schema from DescriptionStore.

        :return:        Container with schema's ids.
        """
        if self._sr is None:
            logger.warning(f"Schema registry is not passed, skipping schema check for {topic}")
            return None
        if schema is None:
            raise ValueError("Couldn't check schema from store.")
        uid = (topic, schema.text, is_key)

        if force is False:
            meta = self._checked.get(uid)
            if meta is not None:
                return meta

        assert schema.type is not None
        meta = self._sr.register_schema(topic, schema.text, schema.type, is_key=is_key)
        self._checked[uid] = meta
        return meta

    # ToDo (tribunsky.kir): Maybe, this method should be public for more simple extension.
    #                       'Template' pattern never works anyway.
    def _encode(self, topic: str, obj: Any, protocol_id: int, is_key: bool = False) -> Optional[bytes]:  # noqa: WPS110
        if obj is None:
            return None

        serializer = self._get_serializer(is_key)
        store = self._get_store(serializer)

        schema = store.get(topic, is_key=is_key)
        if schema is None:
            logger.warning(f"Missing schema for {topic} (key: {is_key}")
            return None
        if schema.empty:
            return serializer.serialize(schema.text, obj, None, topic, is_key=is_key)
        else:
            available_meta = self._check_schema(topic, schema, is_key=is_key)
            # ToDo (tribunsky.kir): `_check_schema()` for now return Optional cause it is used when setting
            #                       producer per topic and should not push schema to on schemaless serializers.
            assert available_meta is not None
            # ToDo (tribunsky.kir): looks like header handler should be also placed per-payload or per-topic,
            #                       because some serializers doesn't use it (e.g. confluent string serializer)
            assert self._header_packer is not None
            # ToDo (tribunsky.kir): check if old client uses
            #                       '{"schema": "{\"type\": \"string\"}"}'
            #                       https://docs.confluent.io/platform/current/schema-registry/develop/using.html#common-sr-api-usage-examples  # noqa: E501
            #                       Looks like the new one doesn't publish string schema at all
            #                       (no type in library for that)
            header = self._header_packer(protocol_id, available_meta)
            return serializer.serialize(schema.text, obj, header, topic, is_key=is_key)

    def _get_store(self, serializer: AbstractSerializer) -> AbstractDescriptionStore:
        serializers_store = getattr(serializer, "store", None)
        if serializers_store is not None:
            return serializers_store
        assert self._store is not None
        return self._store

    def _get_serializer(self, is_key: bool) -> AbstractSerializer:
        return self._key_serializer if is_key else self._value_serializer



from typing import Any, Final, Union, TypeVar, Optional, get_args

try:
    from dataclasses_avroschema.avrodantic import AvroBaseModel
except ImportError:
    from dataclasses_avroschema.pydantic import AvroBaseModel

from pydantic import BaseModel, ConfigDict, create_model
from pydantic_settings import BaseSettings

from wunderkafka.serdes.avromodel.typing import is_generic_type
from wunderkafka.serdes.avromodel.typing.compat import get_generic, is_union_type, create_annotation

A = TypeVar("A", bound=Any)

PYDANTIC_PROTECTED_FIELDS: Final[frozenset[str]] = frozenset(
    {
        "model_config",
        "model_fields",
        # even the latest ones are properties, we don't want to shadow them either
        "model_computed_fields",
        "model_extra",
        "model_fields_set",
    }
)


def exclude_pydantic_class(schema: dict[str, Any]) -> dict[str, Any]:
    schema.pop("pydantic-class", None)
    for value in schema.values():
        if isinstance(value, dict):
            exclude_pydantic_class(value)
        elif isinstance(value, list):
            for item in value:
                if isinstance(item, dict):
                    exclude_pydantic_class(item)
    return schema


def derive_from_pydantic(model_type: type[object]) -> Optional[type[AvroBaseModel]]:
    if issubclass(model_type, AvroBaseModel):
        return model_type
    if issubclass(model_type, BaseModel):
        _check_pydantic_service_fields(model_type)
        return _create_model(model_type)
    return None


def replace_type_in_annotation(annotation: Any) -> Any:
    origin = get_generic(annotation)
    args = get_args(annotation)

    if origin is None:
        if isinstance(annotation, type) and issubclass(annotation, BaseModel):
            return create_model(annotation.__name__, __base__=(annotation, AvroBaseModel))
        else:
            return annotation

    new_args = [replace_type_in_annotation(arg) for arg in args]

    # https://bugs.python.org/issue45418
    if is_union_type(origin):
        origin = Union

    return create_annotation(origin, new_args)


def get_model_attributes(model_type: type[BaseModel]) -> dict[str, Any]:
    attributes: dict[str, Any] = {}
    for field_name, field_info in model_type.model_fields.items():
        # Here we are changing the original model just for schema derivation, so we can override almost everything
        # https://github.com/marcosschroh/dataclasses-avroschema/issues/400
        if field_info.default_factory is not None:
            field_info.default_factory = None
        annotation_type = field_info.annotation
        if isinstance(annotation_type, BaseModel):
            attributes[field_name] = create_model(model_type.__name__, __base__=(annotation_type, AvroBaseModel))
        else:
            if annotation_type is None:
                attributes[field_name] = (annotation_type, field_info)
            else:
                if is_generic_type(annotation_type):
                    new_annotation = replace_type_in_annotation(annotation_type)
                    field_info.annotation = new_annotation
                    attributes[field_name] = (new_annotation, field_info)
                else:
                    if issubclass(annotation_type, BaseModel):
                        new_type = _create_model(annotation_type)
                        field_info.annotation = new_type
                        attributes[field_name] = (new_type, field_info)
                    else:
                        attributes[field_name] = (annotation_type, field_info)
    return attributes


def _check_pydantic_service_fields(model_type: type[object]) -> None:
    if issubclass(model_type, BaseModel):
        all_annotations = set()
        for model in model_type.mro():
            # fragile, maybe it's better to check any of the fields
            is_just_a_base_model = model is BaseModel or model is BaseSettings
            if not is_just_a_base_model:
                for field in vars(model).get("__annotations__", {}):
                    is_real_config_dict = field == "model_config" and model.__annotations__[field] is ConfigDict
                    if not is_real_config_dict:
                        all_annotations.add(field)
        has_protected_fields = all_annotations & PYDANTIC_PROTECTED_FIELDS
        if has_protected_fields:
            msg = " ".join(
                [
                    f"Pydantic model {model_type} has protected fields {has_protected_fields}.",
                    "Please use another name for your field.",
                    "Even if we may derive a schema with such field(s), it would be impossible to instantiate a model",
                ]
            )
            raise ValueError(msg)


def _create_model(model_type: type[BaseModel]) -> type[AvroBaseModel]:
    attributes = get_model_attributes(model_type)
    crafted_model = create_model(model_type.__name__, __base__=(model_type, AvroBaseModel), **attributes)
    assert issubclass(crafted_model, AvroBaseModel)
    return crafted_model



from __future__ import annotations

from enum import Enum
from typing import Final, NamedTuple

from wunderkafka.errors import DeserializerException
from wunderkafka.serdes.structures import Mask, Protocol


class Suffixes(NamedTuple):
    """A pair of suffix strings to append to a Schema Registry's subject for a key and for value."""

    key: str
    value: str


class ProtocolDefinition(NamedTuple):
    """Associates a numeric protocol ID with its Protocol implementation."""

    id: int
    protocol: Protocol


class Vendor(NamedTuple):
    """Holds vendor-specific information for message headers and schema registry."""

    name: str
    suffixes: Suffixes
    protocols: list[ProtocolDefinition]


class VendorRegistry:
    def __init__(self, vendors: list[Vendor]) -> None:
        self.__protocol_id_to_protocol: dict[int, Protocol] = {}
        self.__protocol_id_to_vendor: dict[int, Vendor] = {}

        self.__registered: dict[str, Vendor] = {}
        for vendor in vendors:
            self._register(vendor)

    def get_protocol(self, protocol_id: int) -> Protocol:
        return self.__protocol_id_to_protocol[protocol_id]

    def get_subject_suffixes(self, protocol_id: int) -> Suffixes:
        return self.__protocol_id_to_vendor[protocol_id].suffixes

    def get_protocols(self) -> list[Protocol]:
        return list(self.__protocol_id_to_protocol.values())

    def get_vendors(self) -> list[Vendor]:
        return list(self.__registered.values())

    def _register(self, vendor: Vendor) -> None:
        if vendor.name in self.__registered:
            msg = f"Vendor {vendor!r} already registered"
            raise ValueError(msg)
        for pr in vendor.protocols:
            if pr.id in self.__protocol_id_to_protocol or pr.id in self.__protocol_id_to_vendor:
                msg = f"Protocol ID {pr.id} already registered"
                raise ValueError(msg)
            self.__protocol_id_to_protocol[pr.id] = pr.protocol
            self.__protocol_id_to_vendor[pr.id] = vendor
        self.__registered[vendor.name] = vendor


# see: https://git.io/JvYyC
__VENDOR_REGISTRY: Final[VendorRegistry] = VendorRegistry(
    [
        Vendor(
            name="confluent",
            suffixes=Suffixes("_key", "_value"),
            protocols=[
                # public static final byte CONFLUENT_VERSION_PROTOCOL = 0x0;
                ProtocolDefinition(id=0, protocol=Protocol(4, Mask("I"))),
            ],
        ),
        Vendor(
            name="cloudera",
            suffixes=Suffixes(":k", ""),
            protocols=[
                # public static final byte METADATA_ID_VERSION_PROTOCOL = 0x1;
                ProtocolDefinition(id=1, protocol=Protocol(12, Mask("qi"))),
                # public static final byte VERSION_ID_AS_LONG_PROTOCOL = 0x2;
                ProtocolDefinition(id=2, protocol=Protocol(8, Mask("q"))),
                # public static final byte VERSION_ID_AS_INT_PROTOCOL = 0x3;
                ProtocolDefinition(id=3, protocol=Protocol(4, Mask("I"))),
            ],
        ),
    ]
)


class Actions(str, Enum):
    deserialize = "deserialize"
    serialize = "serialize"


def get_protocol(protocol_id: int, action: Actions) -> Protocol:
    """Retrieve the Protocol implementation for this ID, wrapping errors by action."""
    try:
        return __VENDOR_REGISTRY.get_protocol(protocol_id)
    except KeyError as exc:
        if action == Actions.deserialize:
            msg = f"Unknown protocol extracted from message: {protocol_id}"
            raise DeserializerException(msg) from exc
        msg = f"Unknown protocol_id {protocol_id}!"
        raise ValueError(msg) from exc


def get_subject_suffix(protocol_id: int, *, is_key: bool) -> str:
    """Fetch the correct subject suffix a Schema Registry's subject (key/value)."""
    try:
        suffixes = __VENDOR_REGISTRY.get_subject_suffixes(protocol_id)
    except KeyError as exc:
        # Strictly speaking, if we get unknown protocol_id here,
        # we should raise an exception much earlier (during header unpacking step short after the message was received).
        # But as a Schema Registry client is a separate moving part, we decided to add fuse here
        msg = f"Unknown protocol extracted from message: {protocol_id}"
        raise DeserializerException(msg) from exc
    else:
        return suffixes.key if is_key else suffixes.value


def get_protocols() -> list[Protocol]:
    """Returns a list of all registered header protocols."""
    return __VENDOR_REGISTRY.get_protocols()


def get_vendors() -> list[Vendor]:
    """Returns a list of all registered vendors."""
    return __VENDOR_REGISTRY.get_vendors()



import json
from typing import Any, Union
from pathlib import Path
from urllib.parse import urlparse

from wunderkafka.schema_registry.abc import AbstractHTTPClient


class TestHTTPClient(AbstractHTTPClient):
    def __init__(self, root_dir: Union[str, Path]) -> None:
        self._root = Path(root_dir)

    def make_request(self, relative_url: str, method: str = "GET", body: Any = None, query: Any = None) -> Any:
        # ToDo (tribunsky.kir): rise up actual SR, stop using FS here.
        parsed = urlparse(relative_url)
        filename = self._root / method / f"{parsed.path}.json"
        if parsed.path == "schemas":
            filename = self._root / method / relative_url / body["description"] / f"{parsed.path}.json"
        with open(filename) as fl:
            return json.load(fl)

    def close(self) -> None: ...

    @property
    def base_url(self) -> str:
        return "http://localhost"



from __future__ import annotations

from typing import Protocol, runtime_checkable


# `@runtime_checkable` is needed for pydantic to create schema in StreamResult
@runtime_checkable
class KafkaErrorProtocol(Protocol):
    def code(self) -> int: ...


# `@runtime_checkable` is needed for pydantic to create schema in StreamResult
@runtime_checkable
class MessageProtocol(Protocol):
    """
    Protocol for message-like objects.

    This is used to ensure that the `MessageStub` class can be used in places where a message-like object is expected.
    """

    def error(self) -> None | KafkaErrorProtocol: ...

    def headers(self) -> list[tuple[str, bytes]] | None: ...

    def key(self) -> object: ...

    def offset(self) -> int | None: ...

    def partition(self) -> int | None: ...

    def set_headers(self, headers: list[tuple[str, bytes]]) -> None: ...

    def set_key(self, key: object) -> None: ...

    def set_value(self, value: object) -> None: ...

    def timestamp(self) -> tuple[int, int]: ...

    def topic(self) -> str | None: ...

    def value(self) -> object: ...



######################################################################
########### THIS FILE IS GENERATED, DO NOT EDIT MANUALLY!!! ##########
########### THIS FILE IS GENERATED, DO NOT EDIT MANUALLY!!! ##########
########### THIS FILE IS GENERATED, DO NOT EDIT MANUALLY!!! ##########
########### THIS FILE IS GENERATED, DO NOT EDIT MANUALLY!!! ##########
########### THIS FILE IS GENERATED, DO NOT EDIT MANUALLY!!! ##########
######################################################################

from enum import Enum


class BrokerAddressFamily(str, Enum):
    any = "any"
    v4 = "v4"
    v6 = "v6"


class SecurityProtocol(str, Enum):
    plaintext = "plaintext"
    ssl = "ssl"
    sasl_plaintext = "sasl_plaintext"
    sasl_ssl = "sasl_ssl"


class SslEndpointIdentificationAlgorithm(str, Enum):
    none = "none"
    https = "https"


class SaslOauthbearerMethod(str, Enum):
    default = "default"
    oidc = "oidc"


class ClientDnsLookup(str, Enum):
    use_all_dns_ips = "use_all_dns_ips"
    resolve_canonical_bootstrap_servers_only = "resolve_canonical_bootstrap_servers_only"


class GroupProtocol(str, Enum):
    classic = "classic"
    consumer = "consumer"


class IsolationLevel(str, Enum):
    read_uncommitted = "read_uncommitted"
    read_committed = "read_committed"


class AutoOffsetReset(str, Enum):
    smallest = "smallest"
    earliest = "earliest"
    beginning = "beginning"
    largest = "largest"
    latest = "latest"
    end = "end"
    error = "error"


class CompressionCodec(str, Enum):
    none = "none"
    gzip = "gzip"
    snappy = "snappy"
    lz4 = "lz4"
    zstd = "zstd"


class CompressionType(str, Enum):
    none = "none"
    gzip = "gzip"
    snappy = "snappy"
    lz4 = "lz4"
    zstd = "zstd"


class QueuingStrategy(str, Enum):
    fifo = "fifo"
    lifo = "lifo"



"""
Wunderkafka - the power of librdkafka for humans^W pythons.

Module contains handful re-imports (public API) from the package.

:copyright: (c) 2021 by Severstal Digital.
:license: Apache 2.0, see LICENSE for more details.
"""

# ToDo (tribunsky-kir): API doesn't belong to us, reconsider of exposing confluent_kafka's Message
from confluent_kafka import Message

from wunderkafka.config import ConsumerConfig, ProducerConfig
from wunderkafka.protocols import AnyConsumer, AnyProducer
from wunderkafka.factories.avro import (
    AvroConsumer,
    AvroProducer,
    AvroModelProducer,
    AvroClouderaConsumer,
    AvroConfluentConsumer,
    AvroModelClouderaProducer,
    AvroModelConfluentProducer,
)
from wunderkafka.consumers.bytes import BytesConsumer
from wunderkafka.producers.bytes import BytesProducer
from wunderkafka.config.generated.enums import (
    IsolationLevel,
    AutoOffsetReset,
    CompressionType,
    QueuingStrategy,
    CompressionCodec,
    SecurityProtocol,
    BrokerAddressFamily,
    SslEndpointIdentificationAlgorithm,
)
from wunderkafka.config.schema_registry import SRConfig
from wunderkafka.consumers.subscription import TopicSubscription

__all__ = [
    "Message",
    "ConsumerConfig",
    "ProducerConfig",
    "AnyConsumer",
    "AnyProducer",
    "AvroConsumer",
    "AvroProducer",
    "AvroModelProducer",
    "AvroClouderaConsumer",
    "AvroConfluentConsumer",
    "AvroModelClouderaProducer",
    "AvroModelConfluentProducer",
    "BytesConsumer",
    "BytesProducer",
    "IsolationLevel",
    "AutoOffsetReset",
    "CompressionType",
    "QueuingStrategy",
    "CompressionCodec",
    "SecurityProtocol",
    "BrokerAddressFamily",
    "SslEndpointIdentificationAlgorithm",
    "SRConfig",
    "TopicSubscription",
]



"""This module contains some ready-to-go combinations of the Consumer/Producer."""

from typing import Union, Optional

from wunderkafka import ConsumerConfig, ProducerConfig
from wunderkafka.types import TopicName, MessageDescription
from wunderkafka.structures import SchemaType
from wunderkafka.serdes.avro import FastAvroSerializer, AvroModelSerializer, FastAvroDeserializer
from wunderkafka.serdes.store import AvroModelRepo, SchemaTextRepo
from wunderkafka.serdes.headers import ConfluentClouderaHeadersHandler
from wunderkafka.consumers.bytes import BytesConsumer
from wunderkafka.producers.bytes import BytesProducer
from wunderkafka.schema_registry import SimpleCache, ClouderaSRClient, KerberizableHTTPClient
from wunderkafka.config.krb.rdkafka import config_requires_kerberos
from wunderkafka.consumers.constructor import HighLevelDeserializingConsumer
from wunderkafka.producers.constructor import HighLevelSerializingProducer
from wunderkafka.schema_registry.clients.confluent import ConfluentSRClient


class AvroConsumer(HighLevelDeserializingConsumer):
    """Kafka Consumer client to get AVRO-serialized messages from Confluent/Cloudera installation."""

    def __init__(
        self,
        config: ConsumerConfig,
        *,
        sr_client: Optional[Union[type[ClouderaSRClient], type[ConfluentSRClient]]] = None,
    ) -> None:
        """
        Init consumer from pre-defined blocks.

        :param config:      Configuration for:

                                - Librdkafka consumer.
                                - Schema registry client (conventional options for HTTP).

                            Refer original CONFIGURATION.md (https://git.io/JmgCl) or generated config.

        :param sr_client:   Client for schema registry

        :raises ValueError: If schema registry configuration is missing.
        """
        sr = config.sr
        self._default_timeout: int = 60
        if sr is None:
            raise ValueError(f"Schema registry config is necessary for {self.__class__.__name__}")
        if sr_client is None:
            sr_client = ClouderaSRClient

        super().__init__(
            consumer=BytesConsumer(config),
            schema_registry=sr_client(
                KerberizableHTTPClient(
                    sr, requires_kerberos=config_requires_kerberos(config), cmd_kinit=config.sasl_kerberos_kinit_cmd
                ),
                SimpleCache(),
            ),
            headers_handler=ConfluentClouderaHeadersHandler().parse,
            deserializer=FastAvroDeserializer(),
        )


class AvroProducer(HighLevelSerializingProducer):
    """Kafka Producer client to serialize and send dictionaries or built-in types as messages."""

    def __init__(
        self,
        mapping: Optional[dict[TopicName, MessageDescription]],
        config: ProducerConfig,
        *,
        sr_client: Optional[Union[type[ClouderaSRClient], type[ConfluentSRClient]]] = None,
        protocol_id: int = 1,
    ) -> None:
        """
        Init producer from pre-defined blocks.

        :param mapping:     Topic-to-Schemas mapping.
                            Mapping's value should contain at least message's value schema to be used for serialization.
        :param config:      Configuration for:

                                - Librdkafka producer.
                                - Schema registry client (conventional options for HTTP).

                            Refer original CONFIGURATION.md (https://git.io/JmgCl) or generated config.

        :param sr_client:   Client for schema registry
        :param protocol_id: Protocol id for producer (1 - Cloudera, 0 - Confluent, etc.)

        :raises ValueError: If schema registry configuration is missing.
        """
        sr = config.sr
        if sr is None:
            raise ValueError(f"Schema registry config is necessary for {self.__class__.__name__}")
        if sr_client is None:
            sr_client = ClouderaSRClient
        self._default_timeout: int = 60

        super().__init__(
            producer=BytesProducer(config),
            schema_registry=sr_client(
                KerberizableHTTPClient(
                    sr, requires_kerberos=config_requires_kerberos(config), cmd_kinit=config.sasl_kerberos_kinit_cmd
                ),
                SimpleCache(),
            ),
            header_packer=ConfluentClouderaHeadersHandler().pack,
            serializer=FastAvroSerializer(),
            store=SchemaTextRepo(schema_type=SchemaType.AVRO),
            mapping=mapping,
            protocol_id=protocol_id,
        )


class AvroModelProducer(HighLevelSerializingProducer):
    """Kafka Producer client to serialize and send models or dataclasses as messages."""

    def __init__(
        self,
        mapping: Optional[dict[TopicName, MessageDescription]],
        config: ProducerConfig,
        *,
        sr_client: Optional[Union[type[ClouderaSRClient], type[ConfluentSRClient]]] = None,
        protocol_id: int = 1,
    ) -> None:
        """
        Init producer from pre-defined blocks.

        :param mapping:     Topic-to-Schemas mapping.
                            Mapping's value should contain at least message's value model to derive schema which will
                            be used for serialization.
        :param config:      Configuration for:

                                - Librdkafka producer.
                                - Schema registry client (conventional options for HTTP).

                            Refer original CONFIGURATION.md (https://git.io/JmgCl) or generated config.

        :param sr_client:   Client for schema registry
        :param protocol_id: Protocol id for producer (1 - Cloudera, 0 - Confluent, etc.)

        :raises ValueError: If schema registry configuration is missing.
        """
        sr = config.sr
        if sr is None:
            raise ValueError(f"Schema registry config is necessary for {self.__class__.__name__}")

        if sr_client is None:
            sr_client = ClouderaSRClient
        self._default_timeout: int = 60

        super().__init__(
            producer=BytesProducer(config),
            schema_registry=sr_client(
                KerberizableHTTPClient(
                    sr, requires_kerberos=config_requires_kerberos(config), cmd_kinit=config.sasl_kerberos_kinit_cmd
                ),
                SimpleCache(),
            ),
            header_packer=ConfluentClouderaHeadersHandler().pack,
            serializer=AvroModelSerializer(),
            store=AvroModelRepo(),
            mapping=mapping,
            protocol_id=protocol_id,
        )


class AvroConfluentConsumer(AvroConsumer):
    """Kafka Consumer client to get AVRO-serialized messages from Confluent installation."""

    def __init__(self, config: ConsumerConfig):
        super().__init__(config, sr_client=ConfluentSRClient)


class AvroClouderaConsumer(AvroConsumer):
    """Kafka Consumer client to get AVRO-serialized messages from Cloudera installation."""

    def __init__(self, config: ConsumerConfig):
        super().__init__(config, sr_client=ClouderaSRClient)


class AvroModelConfluentProducer(AvroModelProducer):
    """Kafka Confluent Producer client to serialize and send models or dataclasses as messages."""

    def __init__(self, mapping: Optional[dict[TopicName, MessageDescription]], config: ProducerConfig):
        super().__init__(mapping, config, sr_client=ConfluentSRClient, protocol_id=0)


class AvroModelClouderaProducer(AvroModelProducer):
    """Kafka Cloudera Producer client to serialize and send models or dataclasses as messages."""

    def __init__(self, mapping: Optional[dict[TopicName, MessageDescription]], config: ProducerConfig):
        super().__init__(mapping, config, sr_client=ClouderaSRClient, protocol_id=1)



"""Some timing boilerplate."""

import time
import datetime

from wunderkafka.logger import logger


def now() -> int:
    """
    Return UNIX timestamp in ms.

    :returns:        current timestamp (ms)
    """
    return int(time.time() * 1000)


def ts2dt(ts: float, tz: datetime.tzinfo = datetime.timezone.utc) -> datetime.datetime:
    """
    Convert unix-timestamp to a datetime object.

    This function is primarily intended for formatting timestamps for human readability,
    such as in logs or UI output, rather than for precise timestamp comparisons or decisions.

    :param ts:              Timestamp in seconds or milliseconds.
    :param tz:              Datetime object to use for conversion.
    :returns:               Corresponding datetime object.

    :raises ValueError:     Raised when couldn't coerce float to datetime.
    """
    ts_str_len_ms = 13  # 1559574586123
    ts_str_len_s = 10  # 155957458
    current_ts_len = len(str(int(ts)))
    if current_ts_len <= ts_str_len_s:
        return datetime.datetime.fromtimestamp(ts, tz)
    if current_ts_len == ts_str_len_ms:
        return datetime.datetime.fromtimestamp(round(ts / 1000), tz)
    msg = f"Invalid timestamp: {ts} seems not to be a UNIX-timestamp in seconds or milliseconds"
    raise ValueError(msg)


def ensure_ms_timestamp(ts: float) -> float:
    """
    Ensure timestamp is in milliseconds.

    That's not 100% bulletproof precise method, because, as usual,
    everything related to time is hard and platform-dependent.

    For example,
      - https://stackoverflow.com/a/50860754
      - https://stackoverflow.com/a/32155104

    :param ts:              Timestamp in seconds or milliseconds.
    :returns:               Corresponding timestamp in milliseconds.
    """
    ts_str_len_s = 10  # 155957458
    current_ts_len = len(str(int(ts)))
    if current_ts_len <= ts_str_len_s:
        converted_ts = round(ts * 1000)
        logger.warning(
            f"Timestamp {ts} appears to be in seconds. Converting to milliseconds: {converted_ts:.0f}.",
        )
        return converted_ts
    return ts



from typing import Any, Callable, Optional

from confluent_kafka.serialization import MessageField, SerializationContext
from confluent_kafka.schema_registry import SchemaRegistryClient
from confluent_kafka.schema_registry.json_schema import JSONSerializer as JSONSchemaSerializer

from wunderkafka.serdes.abc import AbstractSerializer, AbstractDescriptionStore

ConfluentAPI = Callable[[Any, SerializationContext], dict]


def any_to_dict(obj: Any, _: SerializationContext) -> dict:
    if isinstance(obj, dict):
        return obj
    return dict(obj)


class JSONSerializer(AbstractSerializer):
    def __init__(
        self,
        schema_registry_client: SchemaRegistryClient,
        store: Optional[AbstractDescriptionStore] = None,
        to_dict: ConfluentAPI = any_to_dict,
    ) -> None:
        self._cache: dict[str, JSONSchemaSerializer] = {}
        self._sr_client = schema_registry_client
        self._to_dict = to_dict
        self.store = store

    def serialize(
        self,
        schema: str,
        obj: Any,
        header: Optional[bytes] = None,
        topic: Optional[str] = None,
        *,
        is_key: Optional[bool] = None,
    ) -> bytes:
        if schema not in self._cache:
            self._cache[schema] = JSONSchemaSerializer(schema, self._sr_client, self._to_dict)
        serializer = self._cache[schema]
        field = MessageField.KEY if is_key else MessageField.VALUE
        return serializer(obj, SerializationContext(topic, field))



import os
import re
import shutil
import logging
import operator
from typing import Union, Optional, NamedTuple
from pathlib import Path
from collections import defaultdict

Name = str
Version = str
Lines = list[str]
Files = dict[Name, Lines]

logging.basicConfig(
    format="%(asctime)s - %(levelname)s - %(message)s",
    level=logging.DEBUG,
)
logger = logging

# mypy cannot into constrained type at the time of on-knee writing of this script
IGNORE_PYDANTIC_TYPE = "# type: ignore[valid-type]"

GROUP_ID = "group.id"
LOG_CB = "log_cb"
SASL_MECHANISMS = "sasl.mechanisms"

CLS_MAPPING = {
    "P": "class RDProducerConfig(RDKafkaConfig):",
    "C": "class RDConsumerConfig(RDKafkaConfig):",
    "*": "class RDKafkaConfig(BaseSettings):",
}

TPL_MAPPING = {
    "P": "PRODUCER_FIELDS = (",
    "C": "CONSUMER_FIELDS = (",
    "*": "COMMON_FIELDS = (",
}

TYPES_MAPPING = {
    "integer": "int",
    "float": "float",
    "string": "str",
    "boolean": "bool",
    "see dedicated API": "Callable",
    "CSV flags": "str",
    "pattern list": "str",
}

DEFAULT_HEADER = [
    "#" * 70,
    "########### THIS FILE IS GENERATED, DO NOT EDIT MANUALLY!!! ##########",
    "########### THIS FILE IS GENERATED, DO NOT EDIT MANUALLY!!! ##########",
    "########### THIS FILE IS GENERATED, DO NOT EDIT MANUALLY!!! ##########",
    "########### THIS FILE IS GENERATED, DO NOT EDIT MANUALLY!!! ##########",
    "########### THIS FILE IS GENERATED, DO NOT EDIT MANUALLY!!! ##########",
    "#" * 70,
    "",
]


class Row(NamedTuple):
    property_name: str
    property_belongs: str
    property_range: str
    property_default: str

    property_importance: str
    property_description: str

    @property
    def comment(self) -> bool:
        return ".." in self.property_range

    @property
    def deprecated(self) -> bool:
        # Markdown
        deprecated = "**deprecated"
        return self.property_description.casefold().startswith(deprecated)

    @property
    def name(self) -> str:
        return self.property_name.replace(".", "_")

    @property
    def cls(self) -> str:
        return CLS_MAPPING[self.property_belongs]

    @property
    def default(self) -> Optional[Union[str, bool, float]]:
        if not self.property_default:
            return None
        if self.type == "boolean":
            return self.property_default.casefold() == "true"
        if self.type == "float":
            return float(self.property_default)
        if self.type in {"string", "CSV flags"}:
            return f"'{self.property_default}'"
        if self.type == "enum value":
            class_name = "".join([word.capitalize() for word in self.property_name.split(".")])
            return f"enums.{class_name}.{self.property_default}"
        return self.property_default

    @property
    def range(self) -> Union[str, tuple[str, str]]:
        delim = ".."
        if delim not in self.property_range:
            # search for bool
            pieces = [piece.strip().casefold() for piece in self.property_range.split(",") if piece.strip()]
            if set(pieces) == {"true", "false"}:
                return "bool"
            logger.warning(f"Couldn't treat range as bool either: {self.property_range}")
            return self.property_range
        pieces = [piece.strip() for piece in self.property_range.split(delim) if piece.strip()]
        if len(pieces) != 2:
            logger.warning(f"Couldn't convert range: {self.property_range}")
            return self.property_range
        ge, le = pieces
        return ge, le

    @property
    def annotation(self) -> str:
        if self.type == "boolean":
            return TYPES_MAPPING[self.type]
        if not self.default:
            return f"Optional[{TYPES_MAPPING[self.type]}]"
        if self.type == "string":
            return TYPES_MAPPING[self.type]
        if self.type == "CSV flags":
            return TYPES_MAPPING[self.type]
        if ".." in self.property_range:
            ge, le = self.range  # type: ignore
            if self.type == "integer":
                return "int"
            elif self.type == "float":
                return "float"
        if self.type == "enum value":
            class_name = "".join([word.capitalize() for word in self.property_name.split(".")])
            return f"enums.{class_name}"
        err_msg = "You missed something, you, stupid! Check: {0} {1} {2}"
        raise ValueError(err_msg.format(self.name, self.property_range, self.default))

    @property
    def type(self) -> str:
        # logger.debug(self.property_description.split('Type: '))
        return self.property_description.split("Type: ")[-1].strip("*")

    def render(self, *, indented: bool = False) -> str:
        if ".." in self.property_range:
            ge, le = self.range  # type: ignore
            return f"    {self.name}: {self.annotation} = Field(ge={ge}, le={le}, default={self.default})"

        if self.property_name == GROUP_ID:
            return f"    {self.name}: str"
        if self.property_name == LOG_CB:
            return f"    logger: Union[Logger, LoggerProtocol, None] = None"
        if self.property_name == SASL_MECHANISMS:
            return f"    # {self.name}: {self.annotation} = {self.default}"
        if self.property_name == "builtin.features":
            lines = ["', '.join(["]
            spaces = " " * (8 + int(indented) * 4)
            for feat in self.default.strip('"').strip("'").split(","):  # type: ignore
                stripped = feat.strip()
                if stripped:
                    lines.append(f"{spaces}'{stripped}',")
            lines.append("    ])")
            return "    {}: {} = {}".format(self.name, self.annotation, "\n".join(lines))
        if self.comment:
            left = f"    {self.name}: {self.annotation} = {self.default}"
            ws = 120 - len(left) - len(IGNORE_PYDANTIC_TYPE) - int(indented) * 4
            if ws < 1:
                ws = 1
            return left + ws * " " + IGNORE_PYDANTIC_TYPE
        else:
            return f"    {self.name}: {self.annotation} = {self.default}"


def read_markdown(filename: Union[str, Path] = "CONFIGURATION.md", *, cut: bool = False) -> list[str]:
    lines = []
    with open(filename) as fl:
        all_lines = [ln.strip() for ln in fl.read().split("\n") if ln.strip()]
        if not cut:
            return all_lines
    for line in all_lines:
        if line.startswith("## Topic configuration properties"):
            break
        lines.append(line)
    return lines


def write_python(lines: Lines, file_name: str) -> None:
    with open(file_name, "w") as fl:
        fl.write("\n".join(lines))


def parse_line(line: str) -> Optional[Row]:
    not_a_rows = ("#", "-", "Property")
    for bad_start in not_a_rows:
        if line.startswith(bad_start):
            return None
    column_data = [piece.strip().replace(r"\|", "|") for piece in line.split(" | ")]
    if len(column_data) != 6:
        logger.warning(f"Couldn't parse: {column_data}")
        return None
    return Row(*column_data)


def parse(lines: list[str], *, allow_deprecated: bool = False) -> list[Row]:
    rows = []
    for line in lines:
        row = parse_line(line)
        if row is not None:
            if row.deprecated:
                if allow_deprecated is False:
                    logger.warning(f"Skipping {row.property_name} as deprecated")
                else:
                    logger.warning(f"{row.property_name} is deprecated")
                    rows.append(row)
            else:
                rows.append(row)
    logger.info(f"Total properties parsed: {len(rows)}")
    rows.append(
        Row(
            property_name="oauth_cb",
            property_belongs="*",
            property_range="",
            property_default="",
            property_importance="",
            property_description="see dedicated API",
        )
    )
    return rows


def group(rows: list[Row]) -> dict[str, list[Row]]:
    grps = defaultdict(list)
    for row in rows:
        grps[row.property_belongs].append(row)
    return grps


def generate_models(groups: dict[str, list[Row]]) -> list[str]:
    properties = DEFAULT_HEADER + [
        "from logging import Logger",
        "from typing import Callable, Optional, Union",
        "",
        "from pydantic import Field",
        "from pydantic_settings import BaseSettings",
        "",
        "# Enums because we can't rely on client code using linters.",
        "# Of course, it will fail with cimpl.KafkaException, but later, when Consumer/Producer are really initiated",
        "from wunderkafka.config.generated import enums",
        "from wunderkafka.config.protocols import LoggerProtocol",
    ]
    already_generated = set()
    for grp in sorted(groups):
        properties.append("")
        properties.append("")
        properties.append(f"{CLS_MAPPING[grp]}")
        pre = []
        uniq = []
        for row in groups[grp]:
            if row.property_name in already_generated:
                logger.warning(f"Skipping generating second field ({row})")
            else:
                if row.property_name == GROUP_ID:
                    pre.append(row)
                else:
                    uniq.append(row)
                already_generated.add(row.property_name)
        properties += [row.render() for row in sorted(pre, key=operator.attrgetter("name"))]
        for prop in sorted(uniq, key=operator.attrgetter("name")):
            if prop.property_name == SASL_MECHANISMS:
                properties.append(
                    "    # ToDo (tribunsky.kir): rethink using aliases? They may need simultaneous validation or may be injected via dict()"
                )
                properties.append(
                    "    # It is just alias, but when setting it manually it may misbehave with current defaults."
                )
            if prop.property_name == LOG_CB:
                properties.append("    # confluent-kafka-python does not use log_cb, it uses logger instead")
            properties.append(prop.render())

    return properties


def generate_fields(groups: dict[str, list[Row]]) -> list[str]:
    properties = DEFAULT_HEADER + [
        "# Why so: not all configuration parameters of librdkafka may be easily replaced from '_' to '.',",
        "#   therefore, we can't convert on-the-fly from  `ssl_ca` without errors",
        "#   and we don't want to have a nice whitelist, which is arguable",
    ]
    for grp in sorted(groups):
        properties.append(f"{TPL_MAPPING[grp]}")
        fields = set()
        for row in groups[grp]:
            if row.property_name == LOG_CB:
                fields.add("logger")
            else:
                fields.add(row.property_name)
        all_fields = sorted(fields)
        properties += [f'    "{field}",' for field in all_fields]
        properties.append(")")
    return properties


def generate_enums(groups: dict[str, list[Row]]) -> list[str]:
    properties = DEFAULT_HEADER + ["from enum import Enum"]
    already_generated = set()
    for grp in sorted(groups):
        for row in groups[grp]:
            if row.type == "enum value":
                if row.property_name in already_generated:
                    logger.warning(f"Skipping second enum for {row.property_name}")
                else:
                    properties.append("")
                    properties.append("")
                    properties += generate_enum(row.property_name, row.property_range)
                    already_generated.add(row.property_name)
    return properties


def urn_to_snake_field(urn: str) -> str:
    tail = urn.rsplit(":")[-1]
    return re.sub(r"[^a-z0-9_]", "", tail.replace("-", "_"))


def generate_enum(prop: str, rng: str) -> list[str]:
    cls_name = "".join([word.capitalize() for word in prop.split(".")])
    header = f"class {cls_name}(str, Enum):"
    fields = []
    for field in rng.split(","):
        if ":" in field:
            field_name = urn_to_snake_field(field)
            line = '    {0} = "{1}"'.format(field_name, field.strip())
        else:
            line = '    {0} = "{0}"'.format(field.strip())
        fields.append(line)
    return [header] + fields


def generate(lines: dict[Version, Files]) -> dict[Version, dict[Name, Lines]]:
    dct: dict[Version, dict[Name, Lines]] = {version: {} for version in lines}

    for libversion in lines:
        for file_name in ("enums.py", "models.py", "fields.py"):
            logger.info(f"Processing {libversion} {file_name}")

            dct[libversion][file_name] = lines[libversion][file_name]
    return dct


def main() -> None:
    lines: dict[Version, Files] = {}
    root_dir = Path(__file__).parent / "versions"
    rev = "1.4.0"
    for sub_path in os.listdir(root_dir):
        path = root_dir / sub_path
        if path.is_dir():
            rev = min(rev, sub_path)
    for sub_path in os.listdir(root_dir):
        path = root_dir / sub_path
        if path.is_dir():
            logger.info(f"{path}: handling...")
            configuration_md = path / "CONFIGURATION.md"
            grouped = group(parse(read_markdown(filename=configuration_md)))
            files = {
                "models.py": generate_models(grouped),
                "fields.py": generate_fields(grouped),
                "enums.py": generate_enums(grouped),
            }
            lines[sub_path] = files
        else:
            logger.info(f"{path}: skipping...")
    versionized_dcts = generate(lines)

    # clean generated dir in order to clean removed versions automatically
    shutil.rmtree("generated/", ignore_errors=True)

    versions_formatted = {}
    for version, dct in versionized_dcts.items():
        version_dir = version_to_dir_name(version)
        version_tuple = tuple(int(part) for part in version.split("."))
        p = Path(f"generated/{version_dir}/")
        p.mkdir(parents=True, exist_ok=True)
        versions_formatted[version_tuple] = version_dir

        for file_name, content in dct.items():
            write_python(content, f"generated/{version_dir}/{file_name}")

    for kind in ("models", "enums", "fields"):
        write_module_file([(k, versions_formatted[k]) for k in sorted(versions_formatted)], kind)


def write_module_file(versions: list[tuple[tuple, str]], kind: str) -> None:
    tmpl = "\nif librdkafka.__version__ >= {version_tuple}:\n    from wunderkafka.config.generated.{version_dir}.{kind} import *  # type: ignore[assignment]"

    file_tmpl = DEFAULT_HEADER + [
        "",
        "from wunderkafka import librdkafka\n",
        "",
    ]
    header_file = "".join(
        tmpl.format(
            version_tuple=version_tpl,
            version_dir=version_dir,
            kind=kind,
        )
        for version_tpl, version_dir in versions
    )
    file_tmpl[-1] = header_file
    with open(f"generated/{kind}.py", "w") as f:
        f.write("\n".join(file_tmpl))


def version_to_dir_name(version: str) -> str:
    return f"v{version.replace('.', '_')}"


def single() -> None:
    grouped = group(parse(read_markdown()))
    models = generate_models(grouped)
    write_python(models, "models.py")
    fields = generate_fields(grouped)
    write_python(fields, "fields.py")
    enums = generate_enums(grouped)
    write_python(enums, "enums.py")


if __name__ == "__main__":
    main()



from typing import Any, Protocol, runtime_checkable


@runtime_checkable
class LoggerProtocol(Protocol):
    def debug(self, msg: str, *args: Any, **kwargs: Any) -> None: ...
    def info(self, msg: str, *args: Any, **kwargs: Any) -> None: ...
    def warning(self, msg: str, *args: Any, **kwargs: Any) -> None: ...
    def error(self, msg: str, *args: Any, **kwargs: Any) -> None: ...
    def critical(self, msg: str, *args: Any, **kwargs: Any) -> None: ...
    def log(self, level: int, msg: str, *args: Any, **kwargs: Any) -> None: ...



from typing import Union, Optional
from pathlib import Path

from pydantic import BaseModel
from pydantic.json_schema import GenerateJsonSchema
from dataclasses_avroschema import AvroModel

from wunderkafka.types import TopicName, KeySchemaDescription, ValueSchemaDescription
from wunderkafka.serdes import avromodel, jsonmodel
from wunderkafka.serdes.abc import AbstractDescriptionStore
from wunderkafka.structures import SchemaType
from wunderkafka.serdes.jsonmodel.derive import JSONClosedModelGenerator


class SchemaTextRepo(AbstractDescriptionStore):
    def __init__(self, schema_type: SchemaType) -> None:
        super().__init__()
        self.schema_type = schema_type

    def add(self, topic: TopicName, value: str, key: str) -> None:
        self._values[topic] = ValueSchemaDescription(text=value, type=self.schema_type)
        if key is not None:
            self._keys[topic] = KeySchemaDescription(text=key, type=self.schema_type)


def _load_from_file(filename: Path) -> str:
    with open(filename) as fl:
        return fl.read().strip()


# FixMe (tribunsky.kir): for now it looks like a crutch, but much better than just `if StringSerializer`
class DummyRepo(AbstractDescriptionStore):
    def add(self, topic: TopicName, value: Union[str, Path], key: Union[str, Path]) -> None:
        # schema = '{"schema": "{\"type\": \"string\"}"}'
        schema = ""
        self._values[topic] = ValueSchemaDescription(text=schema, type=SchemaType.PRIMITIVES)
        if key is not None:
            self._keys[topic] = KeySchemaDescription(text=schema, type=SchemaType.PRIMITIVES)


class StringRepo(DummyRepo): ...


# ToDo (tribunsky.kir): refactor it, maybe add hooks to parent class.
#                       Barbara, forgive us. Looks like AbstractDescriptionStore should be generic.
class SchemaFSRepo(AbstractDescriptionStore):
    def __init__(self, schema_type: SchemaType) -> None:
        super().__init__()
        self.schema_type = schema_type

    def add(self, topic: TopicName, value: Union[str, Path], key: Union[str, Path]) -> None:
        self._values[topic] = ValueSchemaDescription(text=_load_from_file(Path(value)), type=self.schema_type)
        if key is not None:
            self._keys[topic] = KeySchemaDescription(text=_load_from_file(Path(key)), type=self.schema_type)


class AvroModelRepo(AbstractDescriptionStore):
    # ToDo (tribunsky.kir): change Type[AvroModel] to more general alias + check derivation from python built-ins
    def add(self, topic: TopicName, value: type[AvroModel], key: Optional[type[AvroModel]]) -> None:
        self._values[topic] = ValueSchemaDescription(text=avromodel.derive(value, topic), type=SchemaType.AVRO)
        if key is not None:
            self._keys[topic] = KeySchemaDescription(
                text=avromodel.derive(key, topic, is_key=True),
                type=SchemaType.AVRO,
            )


class JSONRepo(AbstractDescriptionStore):
    def add(self, topic: TopicName, value: str, key: Optional[str]) -> None:
        self._values[topic] = ValueSchemaDescription(text=value, type=SchemaType.JSON)
        if key is not None:
            self._keys[topic] = KeySchemaDescription(text=key, type=SchemaType.JSON)


class JSONModelRepo(AbstractDescriptionStore):
    def __init__(self, schema_generator: type[GenerateJsonSchema] = JSONClosedModelGenerator) -> None:
        super().__init__()
        self._schema_generator = schema_generator

    def add(self, topic: TopicName, value: type[BaseModel], key: Optional[type[BaseModel]]) -> None:
        self._values[topic] = ValueSchemaDescription(
            text=jsonmodel.derive(value, self._schema_generator),
            type=SchemaType.JSON,
        )
        if key is not None:
            self._keys[topic] = KeySchemaDescription(
                text=jsonmodel.derive(key, self._schema_generator),
                type=SchemaType.JSON,
            )



from enum import Enum
from typing import Any

from wunderkafka.config.krb.compat import AuthBase

try:
    import requests_kerberos
    from requests_kerberos import HTTPKerberosAuth
except ImportError:
    HAS_KERBEROS = False
else:
    HAS_KERBEROS = True

if HAS_KERBEROS:

    class HTTPKerberosMutualAuth(Enum):
        REQUIRED = requests_kerberos.REQUIRED
        OPTIONAL = requests_kerberos.OPTIONAL
        DISABLED = requests_kerberos.DISABLED

else:

    class HTTPKerberosMutualAuth(Enum):  # type: ignore[no-redef]
        REQUIRED = 1
        OPTIONAL = 2
        DISABLED = 3

    class HTTPKerberosAuth(AuthBase):  # type: ignore[no-redef] # noqa: F811
        def __init__(self, *args: Any, **kwargs: Any) -> None:
            message = " ".join(
                [
                    "Something went wrong: trying to use HTTPKerberosAuth while missing requests-kerberos.",
                    "Maybe it is unexpected manual usage. Please, install requests-kerberos.",
                ]
            )
            raise ImportError(message)



"""This module contains subscription helpers."""

from __future__ import annotations

import datetime

from confluent_kafka import OFFSET_END, OFFSET_BEGINNING

from wunderkafka.time import ensure_ms_timestamp
from wunderkafka.structures import Offset, Timestamp


def choose_offset(
    *,
    from_beginning: bool | None = None,
    offset: int | None = None,
    ts: float | None = None,
    dt: datetime.datetime | None = None,
    with_timedelta: datetime.timedelta | None = None,
) -> Offset | Timestamp | None:
    """
    Choose subscription method (Offset, Timestamp, None), corresponding to given values.

    :param from_beginning:  If a flag is set, return specific offset corresponding to the beginning /end of the topic.
    :param offset:          If set, will return an Offset object.
    :param ts:              If set, will return Timestamp.
    :param dt:              If set, will get Timestamp from a datetime object.
    :param with_timedelta:  If set, will calculate timestamp for corresponding timedelta.

    :raises ValueError:     If multiple values are set: there is no sense in attempt to subscribe to a topic via
                            multiple offsets (excluding a multiple partitions scenario, which is not currently
                            supported).

    :return:                None if no values are specified, Offset or Timestamp otherwise.
    """
    args = (from_beginning, offset, ts, with_timedelta)
    if sum([arg is not None for arg in args]) > 1:
        msg = "Only one subscription method per time is allowed!"
        raise ValueError(msg)
    if from_beginning is not None:
        offset = OFFSET_BEGINNING if from_beginning else OFFSET_END
    if offset is not None:
        return Offset(offset)
    if with_timedelta is not None:
        start = datetime.datetime.now(datetime.timezone.utc) - with_timedelta
        ts = start.timestamp() * 1000
    if dt is not None:
        ts = dt.timestamp() * 1000
    if ts is not None:
        return Timestamp(int(ensure_ms_timestamp(ts)))
    return None


class TopicSubscription:
    """Allows custom definition of subscription for a topic without the need to build a full TopicPartition list."""

    # # TODO (tribunsky.kir): reconsider API of 'how'
    #                         https://github.com/severstal-digital/wunderkafka/issues/89
    def __init__(  # noqa: PLR0913
        self,
        topic: str,
        *,
        from_beginning: bool | None = None,
        offset: int | None = None,
        ts: float | None = None,
        dt: datetime.datetime | None = None,
        with_timedelta: datetime.timedelta | None = None,
    ) -> None:
        """
        Init topic subscription object.

        Only one method of subscription per topic is allowed at a time:

        - From the beginning (depends on your retention policy)
        - from the end (consume only latest messages)
        - via specific offset
        - via specific timestamp
        - via specific datetime (datetime.datetime object)
        - via specific timedelta (from current datetime)
        - no special option (consumer will use "default" value of auto.offset.reset)

        :param topic:           Topic to subscribe.
        :param from_beginning:  If True, subscribe to get the earliest available messages.
                                If False, get the latest messages.
        :param offset:          Subscribe to specific offset.
                                If offset not found, it will behave with built-in default.
        :param ts:              Subscribe to specific timestamp (milliseconds).
                                If timestamp is not found, will behave with built-in default.
        :param dt:              Subscribe to specific datetime.
                                If datetime is not found, will behave with built-in default.
        :param with_timedelta:  Subscribe to some moments in the past, from current datetime for a given timedelta.
                                Will calculate specific timestamp and subscribe via ts.
        """
        self.topic = topic
        self.how = choose_offset(
            from_beginning=from_beginning,
            offset=offset,
            ts=ts,
            dt=dt,
            with_timedelta=with_timedelta,
        )



import inspect
from typing import Any, Annotated, get_origin

# We check it via import to avoid using nested imports in implementation in `is_union_type() function`
HAS_UNION_TYPE = True
try:
    from types import UnionType  # type: ignore[attr-defined]
except ImportError:
    HAS_UNION_TYPE = False


def create_annotation(generic: Any, types_list: list[type[object]]) -> type[object]:
    # return generic[Union[types_list]]
    return generic[tuple(types_list)]


# Same as `sys.version_info <= (3, 10)`
if HAS_UNION_TYPE is False:

    def is_union_type(generic: Any) -> bool:
        return False
else:

    def is_union_type(generic: Any) -> bool:
        return inspect.isclass(generic) and issubclass(generic, UnionType)


def get_generic(annotation: Any) -> Any:
    return get_origin(annotation)


def is_annotated_type(annotation: Any) -> bool:
    return get_origin(annotation) is Annotated



from __future__ import annotations

from confluent_kafka import libversion


def get_librdkafka_version_tuple(version: str | tuple[str, int]) -> tuple[int, ...]:
    if not isinstance(version, str):
        _triplet, _ = version
    else:
        _triplet = version
    if "-" in _triplet:
        _triplet = _triplet.split("-")[0]
    return tuple(int(digit) for digit in _triplet.split("."))


__version__ = get_librdkafka_version_tuple(libversion())



"""This module contains some ready-to-go combinations of the Consumer/Producer."""

from typing import Optional

from wunderkafka import BytesConsumer, BytesProducer, ConsumerConfig, ProducerConfig
from wunderkafka.types import TopicName, MessageDescription
from wunderkafka.serdes.avro import FastAvroDeserializer
from wunderkafka.serdes.store import AvroModelRepo
from wunderkafka.serdes.headers import ConfluentClouderaHeadersHandler
from wunderkafka.schema_registry import SimpleCache, ConfluentSRClient, KerberizableHTTPClient
from wunderkafka.config.krb.rdkafka import config_requires_kerberos
from wunderkafka.consumers.constructor import HighLevelDeserializingConsumer
from wunderkafka.producers.constructor import HighLevelSerializingProducer
from wunderkafka.serdes.json.deserializers import JSONDeserializer
from wunderkafka.serdes.avromodel.serializers import AvroModelSerializer
from wunderkafka.serdes.jsonmodel.serializers import JSONModelSerializer
from wunderkafka.serdes.schemaless.string.serializers import StringSerializer
from wunderkafka.serdes.schemaless.string.deserializers import StringDeserializer


class AvroStringConsumer(HighLevelDeserializingConsumer):
    """Kafka Consumer client to get messages with avro-serialized values and string-serialized keys."""

    def __init__(
        self,
        config: ConsumerConfig,
        *,
        sr_client: Optional[type[ConfluentSRClient]] = None,
    ) -> None:
        """
        Init consumer from pre-defined blocks.

        :param config:      Configuration for:

                                - Librdkafka consumer.
                                - Schema registry client (conventional options for HTTP).

                            Refer original CONFIGURATION.md (https://git.io/JmgCl) or generated config.

        :param sr_client:   Client for schema registry

        :raises ValueError: If schema registry configuration is missing.
        """
        self._default_timeout: int = 60
        sr = config.sr
        if sr is None:
            raise ValueError(f"Schema registry config is necessary for {self.__class__.__name__}")
        if sr_client is None:
            sr_client = ConfluentSRClient

        super().__init__(
            consumer=BytesConsumer(config),
            schema_registry=sr_client(
                KerberizableHTTPClient(sr),
                SimpleCache(),
            ),
            headers_handler=ConfluentClouderaHeadersHandler().parse,
            value_deserializer=FastAvroDeserializer(),
            key_deserializer=StringDeserializer(),
            stream_result=True,
        )


class AvroModelStringProducer(HighLevelSerializingProducer):
    """Kafka Producer client to send models as avro-serialized message values and string-serialized keys."""

    def __init__(
        self,
        mapping: Optional[dict[TopicName, MessageDescription]],
        config: ProducerConfig,
        *,
        sr_client: Optional[type[ConfluentSRClient]] = None,
    ) -> None:
        """
        Init producer from pre-defined blocks.

        :param mapping:     Topic-to-Schemas mapping.
                            Mapping's value should contain at least message's value model to derive schema which will
                            be used for serialization.
        :param config:      Configuration for:

                                - Librdkafka producer.
                                - Schema registry client (conventional options for HTTP).

                            Refer original CONFIGURATION.md (https://git.io/JmgCl) or generated config.

        :param sr_client:   Client for schema registry

        :raises ValueError: If schema registry configuration is missing.
        """
        self._default_timeout: int = 60
        sr = config.sr
        if sr is None:
            raise ValueError(f"Schema registry config is necessary for {self.__class__.__name__}")

        if sr_client is None:
            sr_client = ConfluentSRClient

        super().__init__(
            producer=BytesProducer(config),
            schema_registry=sr_client(
                KerberizableHTTPClient(
                    sr,
                    requires_kerberos=config_requires_kerberos(config),
                    cmd_kinit=config.sasl_kerberos_kinit_cmd,
                ),
                SimpleCache(),
            ),
            header_packer=ConfluentClouderaHeadersHandler().pack,
            value_serializer=AvroModelSerializer(AvroModelRepo()),
            key_serializer=StringSerializer(),
            mapping=mapping,
            protocol_id=0,
        )


class JSONStringConsumer(HighLevelDeserializingConsumer):
    """Kafka Consumer client to get messages with JSON-serialized values and string-serialized keys."""

    def __init__(
        self,
        config: ConsumerConfig,
        *,
        sr_client: Optional[type[ConfluentSRClient]] = None,
    ) -> None:
        """
        Init consumer from pre-defined blocks.

        :param config:      Configuration for:

                                - Librdkafka consumer.
                                - Schema registry client (conventional options for HTTP).

                            Refer original CONFIGURATION.md (https://git.io/JmgCl) or generated config.

        :param sr_client:   Client for schema registry

        :raises ValueError: If schema registry configuration is missing.
        """
        self._default_timeout: int = 60
        sr = config.sr
        if sr is None:
            raise ValueError(f"Schema registry config is necessary for {self.__class__.__name__}")
        if sr_client is None:
            sr_client = ConfluentSRClient

        super().__init__(
            consumer=BytesConsumer(config),
            schema_registry=sr_client(
                KerberizableHTTPClient(sr),
                SimpleCache(),
            ),
            headers_handler=ConfluentClouderaHeadersHandler().parse,
            value_deserializer=JSONDeserializer(),
            key_deserializer=StringDeserializer(),
            stream_result=True,
        )


class JSONModelStringProducer(HighLevelSerializingProducer):
    """Kafka Producer client to send models as JSON-serialized message values and string-serialized keys."""

    def __init__(
        self,
        mapping: Optional[dict[TopicName, MessageDescription]],
        config: ProducerConfig,
        *,
        sr_client: Optional[type[ConfluentSRClient]] = None,
    ) -> None:
        """
        Init producer from pre-defined blocks.

        :param mapping:     Topic-to-Schemas mapping.
                            Mapping's value should contain at least message's value model to derive schema which will
                            be used for serialization.
        :param config:      Configuration for:

                                - Librdkafka producer.
                                - Schema registry client (conventional options for HTTP).

                            Refer original CONFIGURATION.md (https://git.io/JmgCl) or generated config.

        :param sr_client:   Client for schema registry

        :raises ValueError: If schema registry configuration is missing.
        """
        self._default_timeout: int = 60

        sr = config.sr
        if sr is None:
            raise ValueError(f"Schema registry config is necessary for {self.__class__.__name__}")

        if sr_client is None:
            sr_client = ConfluentSRClient

        schema_registry = sr_client(
            KerberizableHTTPClient(
                sr,
                requires_kerberos=config_requires_kerberos(config),
                cmd_kinit=config.sasl_kerberos_kinit_cmd,
            ),
            SimpleCache(),
        )
        super().__init__(
            producer=BytesProducer(config),
            schema_registry=schema_registry,
            header_packer=ConfluentClouderaHeadersHandler().pack,
            value_serializer=JSONModelSerializer(schema_registry.client),
            key_serializer=StringSerializer(),
            mapping=mapping,
            protocol_id=0,
        )



from wunderkafka.serdes.avro.serializers import FastAvroSerializer
from wunderkafka.serdes.avro.deserializers import FastAvroDeserializer
from wunderkafka.serdes.avromodel.serializers import AvroModelSerializer

__all__ = [
    "FastAvroSerializer",
    "FastAvroDeserializer",
    "AvroModelSerializer",
]



import re
import json
from typing import Any
from dataclasses import is_dataclass

from dataclasses_avroschema import AvroModel

from wunderkafka.logger import logger
from wunderkafka.serdes.avromodel.pydantic import derive_from_pydantic, exclude_pydantic_class


def derive(model_type: type[object], topic: str, *, is_key: bool = False) -> str:
    pydantic_model = derive_from_pydantic(model_type)
    if pydantic_model is None:
        if is_dataclass(model_type):
            # https://github.com/python/mypy/issues/14941
            model_schema = model_type.avro_schema_to_python()  # type: ignore
        else:
            # Non-dataclasses objects may allow mixing defaults and non-default fields order,
            # so to still reuse dataclasses_avroschema, we extract fields and reorder them to satisfy dataclasses
            # restrictions, than reorder them back.
            # All these hacks will work only for flat schemas:
            #   we avoid describing objects via nested types for HDFS's sake.
            attributes = _extract_attributes(model_type)
            ordering = list(attributes["__annotations__"])
            crafted_model: type[AvroModel] = type(model_type.__name__, (AvroModel,), attributes)
            model_schema = crafted_model.avro_schema_to_python()
            fields_map = {field_data["name"]: field_data for field_data in model_schema["fields"]}
            reordered_fields = [fields_map[attr] for attr in ordering]
            model_schema["fields"] = reordered_fields
    else:
        model_schema = pydantic_model.avro_schema_to_python()
    model_schema.pop("doc", None)
    suffix = "key" if is_key else "value"
    if hasattr(model_type, "Meta") and hasattr(model_type.Meta, "name"):
        model_schema["name"] = model_type.Meta.name
    else:
        model_schema["name"] = "{}_{}".format(re.sub(r"[^\w_]", "_", topic), suffix)
    if model_schema["name"].startswith("_"):
        logger.warning("Topic name {0} starts with underscore, which is not allowed for Avro schema names.")
        model_schema["name"] = model_schema["name"][1:]
    if pydantic_model:
        # I would not prefer to just override PydanticParser because it always may be broken or thrown-out
        model_schema = exclude_pydantic_class(model_schema)
    for field in model_schema.get("fields", []):
        field.pop("doc", None)
    return json.dumps(model_schema)


def _extract_attributes(type_: type[object]) -> dict[str, Any]:
    fields = vars(type_).get("__annotations__", {})
    _, *parents = type_.mro()
    for base in parents:
        new_fields = {**vars(base).get("__annotations__", {})}
        # Currently it is `model_config: ClassVar[SettingsConfigDict]`
        # https://github.com/pydantic/pydantic-settings/blob/919a20b77527ecc1cd6eeb0a09ca22cc21486fb8/pydantic_settings/main.py#L166
        for field in ["__slots__", "klass", "metadata", "schema_def", "__config__", "model_config"]:
            new_fields.pop(field, None)
        fields = {**new_fields, **fields}

    attributes = {}
    attributes.update({"__annotations__": fields})
    meta = vars(type_).get("Meta", None)
    if meta is not None:
        attributes.update({"Meta": meta})
    return attributes



import sys

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec


__all__ = [
    "ParamSpec",
    "Self",
]



import json
from typing import Any, Optional

from confluent_kafka.schema_registry.json_schema import JSONDeserializer as JSONDSchemaDeserializer

from wunderkafka.serdes.abc import AbstractDeserializer


class SchemaLessJSONDeserializer(AbstractDeserializer):
    def __init__(self) -> None:
        self.schemaless = True
        self._cache: dict[str, JSONDSchemaDeserializer] = {}

    def deserialize(self, schema: str, blob: bytes, seek_pos: Optional[int] = None) -> Any:
        return json.loads(blob)



"""
Module contains interface-like skeletons for producer.

- Inherited from confluent-kafka Producer to use it as nested entity
- high-level producer which is able to handle message's schemas
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TypeVar

from confluent_kafka import Producer

from wunderkafka.types import MsgKey, MsgValue, DeliveryCallback
from wunderkafka.compat import ParamSpec

P = ParamSpec("P")
T = TypeVar("T")


class AbstractProducer(Producer):
    """Extension point for the original Producer API."""

    # TODO (tribunsky.kir): rethink API?
    #                       https://github.com/severstal-digital/wunderkafka/issues/91
    # https://github.com/python/mypy/issues/13966
    def send_message(  # noqa: PLR0913
        self,
        topic: str,
        value: str | bytes | None = None,
        key: str | bytes | None = None,
        partition: int | None = None,
        on_delivery: DeliveryCallback | None = None,
        *args: P.args,  # type: ignore[valid-type]
        blocking: bool = False,
        **kwargs: P.kwargs,  # type: ignore[valid-type]
    ) -> None:
        """
        Send an encoded message to Kafka almost immediately.

        This method overlaps the original producers' method.

        :param topic:           Target topic against which we are working for this call.
        :param value:           Message's value encoded object to be sent.
        :param key:             Message's key encoded object to be sent.
        :param partition:       Target partition to produce to. If none, uses configured built-in
                                partitioner by default.
        :param on_delivery:     Callback to be executed on successful/failed delivery.
        :param args:            Other positional arguments to call with original produce() API.
        :param blocking:        If True, will block execution until a delivery result is retrieved
                                (calls flush() under the hood). Otherwise, will produce a message to
                                builtin queue and call poll() to trigger producer's events.
        :param kwargs:          Other keyword arguments to call with original produce() API.
        """
        raise NotImplementedError


class AbstractSerializingProducer(ABC):
    """High-level interface for extended producer."""

    @abstractmethod
    # https://github.com/python/mypy/issues/13966
    def send_message(  # noqa: PLR0913
        self,
        topic: str,
        value: MsgValue = None,
        key: MsgKey = None,
        partition: int | None = None,
        on_delivery: DeliveryCallback | None = None,
        *args: P.args,  # type: ignore[valid-type]
        blocking: bool = False,
        **kwargs: P.kwargs,  # type: ignore[valid-type]
    ) -> None:
        """
        Encode and send a message to Kafka.

        This method overlaps AbstractProducer's method and intended to use produce() of real nested producer.

        :param topic:           Target topic against which we are working for this call.
        :param value:           Message's value object to be encoded and sent.
        :param key:             Message's key object to be encoded and sent.
        :param partition:       Target partition to produce to. If none, uses configured built-in
                                partitioner by default.
        :param on_delivery:     Callback to be executed on successful/failed delivery.
        :param args:            Other positional arguments to call with original produce() API.
        :param blocking:        If True, will block execution until a delivery result is retrieved
                                (calls flush() under the hood). Otherwise, will produce a message to
                                builtin queue and call poll() to trigger producer's events.
        :param kwargs:          Other keyword arguments to call with original produce() API.
        """

    # TODO (tribunsky.kir): change naming. It is like subscribe, but publish/add_topic whatever
    #                       https://github.com/severstal-digital/wunderkafka/issues/92
    @abstractmethod
    def set_target_topic(self, topic: str, value: T, key: T | None = None, *, lazy: bool = False) -> None:
        """
        Make producer aware how it should work with a specific topic.

        This method is used to instantiate producer with mapping, but also adds schemas in runtime.

        :param topic:           Target topic to specify schema description against.
        :param value:           Message's value schema description.
        :param key:             Message's key schema description.
        :param lazy:            If True, do not register schema in registry during __init__,
                                or before the first attempt to send a message for a given topic.
                                Otherwise, register schema immediately.
        """

    @abstractmethod
    def flush(self, timeout: float) -> int:
        """
        Wait for all messages in the Producer queue to be delivered.

        This method overlaps the original producers' method and will use the nested producer.

        :param timeout:         Maximum time to block.
        :return:                Number of messages still in queue.
        """



"""This module contains common datastructures which is used in the package."""

from __future__ import annotations

import datetime
from enum import Enum
from dataclasses import dataclass

from wunderkafka.time import ts2dt
from wunderkafka.serdes.vendors import get_subject_suffix


class SchemaType(str, Enum):
    AVRO = "AVRO"
    JSON = "JSON"
    PROTOBUF = "PROTOBUF"
    # Strings, doubles, integers
    PRIMITIVES = "PRIMITIVES"


@dataclass(frozen=True)
class Timestamp:
    value: int

    def __str__(self) -> str:
        tz_local = datetime.datetime.now().astimezone().tzinfo
        return f"{self.__class__.__name__}: {self.value:.2f} ({ts2dt(self.value).astimezone(tz_local)})"


@dataclass(frozen=True)
class Offset:
    value: int


# TODO (tribunsky.kir): compose header & meta in symmetrical way
#                       https://github.com/severstal-digital/wunderkafka/issues/88
@dataclass(frozen=True)
class ParsedHeader:
    protocol_id: int
    meta_id: int | None
    schema_id: int | None
    schema_version: int | None
    size: int


@dataclass(frozen=True)
class SchemaMeta:
    topic: str
    is_key: bool
    header: ParsedHeader

    @property
    def subject(self) -> str:
        """
        Return topic mapped to schema registry entity.

        :return:            String which should be used as a path in schema registry's url
                            corresponding to a given topic.
        """
        suffix = get_subject_suffix(self.header.protocol_id, is_key=self.is_key)
        return f"{self.topic}{suffix}"


# TODO (tribunsky.kir): add cross-validation of invariants on the model itself?
#                       This awkward class may be gone with:
#                       https://github.com/severstal-digital/wunderkafka/issues/88
@dataclass(frozen=True)
class SRMeta:
    """Meta, which is retrieved after schema registration."""

    # Confluent always has schema_id, but one of cloudera protocols doesn't use it
    schema_id: int | None
    # Confluent has schema_version, but serdes works around schema_id, which is a unique identifier there.
    schema_version: int | None
    # Confluent doesn't have metaId
    meta_id: int | None = None


@dataclass(frozen=True)
class SerializerSchemaDescription:
    """
    Class to allow a contract extension between moving parts of serialization.

    Usually schema is represented by text, but separate class adds abstraction.
    """

    text: str

    @property
    def empty(self) -> bool:
        """
        Return True if schema is empty.

        Some serializers (e.g., confluent StringSerializer) practically don't have an effective schema.
        """
        return self.text == ""


@dataclass(frozen=True)
class DeserializerSchemaDescription(SerializerSchemaDescription):
    """
    Class to allow a contract extension between moving parts of deserialization.

    Usually schema is represented by text, but separate class adds abstraction.
    """

    text: str
    type: SchemaType
